from __future__ import annotations

import os

import procrunner
import pytest

import scitbx.matrix
from cctbx import sgtbx
from cctbx.sgtbx.lattice_symmetry import metric_subgroups
from dxtbx.model import Crystal, Experiment, ExperimentList
from dxtbx.serialize import load

import dials.command_line.reindex
from dials.array_family import flex
from dials.command_line.reindex import reindex_experiments


def test_reindex(dials_regression, tmpdir):
    data_dir = os.path.join(dials_regression, "indexing_test_data", "i04_weak_data")
    pickle_path = os.path.join(data_dir, "indexed.pickle")
    experiments_path = os.path.join(data_dir, "experiments.json")
    commands = [
        "dials.reindex",
        pickle_path,
        experiments_path,
        "change_of_basis_op=2a,b,c",
        "space_group=P1",
    ]
    result = procrunner.run(commands, working_directory=tmpdir)
    assert not result.returncode and not result.stderr

    old_reflections = flex.reflection_table.from_file(pickle_path)
    assert tmpdir.join("reindexed.refl").check()
    new_reflections = flex.reflection_table.from_file(
        tmpdir.join("reindexed.refl").strpath
    )
    old_experiments = load.experiment_list(experiments_path, check_format=False)
    assert tmpdir.join("reindexed.expt").check()
    new_experiments = load.experiment_list(
        tmpdir.join("reindexed.expt").strpath, check_format=False
    )
    h1, k1, l1 = old_reflections["miller_index"].as_vec3_double().parts()
    h2, k2, l2 = new_reflections["miller_index"].as_vec3_double().parts()
    assert 2 * h1 == pytest.approx(h2)
    assert k1 == pytest.approx(k2)
    assert l1 == pytest.approx(l2)
    old_uc_params = old_experiments[0].crystal.get_unit_cell().parameters()
    new_uc_params = new_experiments[0].crystal.get_unit_cell().parameters()
    assert new_uc_params[0] == pytest.approx(2 * old_uc_params[0])
    assert new_uc_params[1:] == pytest.approx(old_uc_params[1:])
    assert old_experiments[0].crystal.get_space_group().type().hall_symbol() == " P 1"
    assert new_experiments[0].crystal.get_space_group().type().hall_symbol() == " P 1"

    # set space group P4
    cb_op = sgtbx.change_of_basis_op("a,b,c")
    commands = [
        "dials.reindex",
        experiments_path,
        "space_group=P4",
        f"change_of_basis_op={cb_op}",
        "output.experiments=P4.expt",
    ]
    result = procrunner.run(commands, working_directory=tmpdir)
    assert not result.returncode and not result.stderr
    assert tmpdir.join("P4.expt").check()

    # apply one of the symops from the space group
    cb_op = sgtbx.change_of_basis_op("-x,-y,z")
    commands = [
        "dials.reindex",
        "P4.expt",
        f"change_of_basis_op={cb_op}",
        "output.experiments=P4_reindexed.expt",
    ]
    result = procrunner.run(commands, working_directory=tmpdir)
    assert not result.returncode and not result.stderr
    assert tmpdir.join("P4_reindexed.expt").check()

    new_experiments1 = load.experiment_list(
        tmpdir.join("P4_reindexed.expt").strpath, check_format=False
    )
    assert new_experiments1[0].crystal.get_A() == pytest.approx(
        old_experiments[0].crystal.change_basis(cb_op).get_A(), abs=1e-5
    )

    cb_op = sgtbx.change_of_basis_op("-x,-y,z")
    commands = [
        "dials.reindex",
        "P4.expt",
        "change_of_basis_op=auto",
        "reference.experiments=P4_reindexed.expt",
        "output.experiments=P4_reindexed2.expt",
    ]
    result = procrunner.run(commands, working_directory=tmpdir)
    assert not result.returncode and not result.stderr
    new_experiments2 = load.experiment_list(
        tmpdir.join("P4_reindexed2.expt").strpath, check_format=False
    )
    assert new_experiments1[0].crystal.get_A() == pytest.approx(
        new_experiments2[0].crystal.get_A()
    )

    # verify that the file can be read by dials.show
    commands = [
        "dials.show",
        tmpdir.join("reindexed.refl").strpath,
    ]
    result = procrunner.run(commands, working_directory=tmpdir)
    assert not result.returncode and not result.stderr


def test_reindex_multi_sequence(dials_regression, tmpdir):
    data_dir = os.path.join(dials_regression, "indexing_test_data", "multi_sweep")
    pickle_path = os.path.join(data_dir, "indexed.pickle")
    experiments_path = os.path.join(data_dir, "experiments.json")
    commands = [
        "dials.reindex",
        pickle_path,
        experiments_path,
        "change_of_basis_op=x+y,x-z,y-z",
    ]

    result = procrunner.run(commands, working_directory=tmpdir)
    assert not result.returncode and not result.stderr
    assert tmpdir.join("reindexed.refl").check()
    assert tmpdir.join("reindexed.expt").check()

    old_reflections = flex.reflection_table.from_file(pickle_path)
    new_reflections = flex.reflection_table.from_file(
        tmpdir.join("reindexed.refl").strpath
    )
    assert len(old_reflections) == len(new_reflections)
    new_experiments = load.experiment_list(
        tmpdir.join("reindexed.expt").strpath, check_format=False
    )
    new_cs = new_experiments[0].crystal.get_crystal_symmetry()
    assert new_cs.unit_cell().parameters() == pytest.approx(
        (
            6.189939294071243,
            6.189939294071243,
            6.189939294071242,
            113.16417286469935,
            107.65690626466579,
            107.65690626466579,
        )
    )
    assert (
        new_experiments[0].crystal.get_space_group().type().hall_symbol()
        == " I 4 (x+y,y+z,x+z)"
    )


def test_reindex_against_reference(dials_regression, tmpdir):
    """Test the reindexing against a reference dataset functionality."""
    data_dir = os.path.join(dials_regression, "indexing_test_data", "i04_weak_data")
    pickle_path = os.path.join(data_dir, "indexed.pickle")
    experiments_path = os.path.join(data_dir, "experiments.json")

    commands = [
        "dials.reindex",
        pickle_path,
        experiments_path,
        "change_of_basis_op=a,b,c",
        "space_group=P4",
        "output.reflections=P4.refl",
        "output.experiments=P4.expt",
    ]

    result = procrunner.run(commands, working_directory=tmpdir)
    assert not result.returncode and not result.stderr
    assert os.path.exists(tmpdir / "P4.refl")
    assert os.path.exists(tmpdir / "P4.expt")
    new_experiments = load.experiment_list(
        tmpdir.join("P4.expt").strpath, check_format=False
    )
    assert new_experiments[0].crystal.get_space_group().type().hall_symbol() == " P 4"

    # Now have something in P4, get another dataset in a different indexing scheme

    cb_op = sgtbx.change_of_basis_op("a,-b,-c")
    commands = [
        "dials.reindex",
        "P4.refl",
        "P4.expt",
        f"change_of_basis_op={cb_op}",
        "output.experiments=P4_reindexed.expt",
        "output.reflections=P4_reindexed.refl",
    ]
    result = procrunner.run(commands, working_directory=tmpdir)
    assert not result.returncode and not result.stderr

    # now run reference reindexing
    commands = [
        "dials.reindex",
        "P4.refl",
        "P4.expt",
        "reference.experiments=P4_reindexed.expt",
        "reference.reflections=P4_reindexed.refl",
    ]
    result = procrunner.run(commands, working_directory=tmpdir)
    assert not result.returncode and not result.stderr

    # expect reindexed_reflections to be same as P4_reindexed, not P4_reflections
    reindexed_reflections = flex.reflection_table.from_file(
        tmpdir.join("reindexed.refl").strpath
    )
    P4_reindexed = flex.reflection_table.from_file(
        tmpdir.join("P4_reindexed.refl").strpath
    )
    P4_reflections = flex.reflection_table.from_file(tmpdir.join("P4.refl").strpath)

    h1, k1, l1 = reindexed_reflections["miller_index"].as_vec3_double().parts()
    h2, k2, l2 = P4_reindexed["miller_index"].as_vec3_double().parts()
    h3, k3, l3 = P4_reflections["miller_index"].as_vec3_double().parts()

    # hkl1 and hkl2 should be same, as should have been reindexed by against the
    # reference, with the program determining a reindexing operator of a,-b,-c
    assert list(h1) == pytest.approx(list(h2))
    assert list(l1) == pytest.approx(list(l2))
    assert list(k1) == pytest.approx(list(k2))
    # h1 and h3 should be same, but not l and k, as these dataset should differ
    # by an a twinning operator of a,-b,-c
    assert list(h1) == pytest.approx(list(h3))
    assert list(l1) != pytest.approx(list(l3))
    assert list(k1) != pytest.approx(list(k3))


def test_reindex_experiments():
    # See also https://github.com/cctbx/cctbx_project/issues/424
    cs = sgtbx.space_group_info("I23").any_compatible_crystal_symmetry(volume=100000)
    B = scitbx.matrix.sqr(cs.unit_cell().fractionalization_matrix()).transpose()
    cryst = Crystal(B, cs.space_group())
    n_scan_points = 10
    A_at_scan_points = [(1, 0, 0, 0, 1, 0, 0, 0, 1)] * n_scan_points
    cryst.set_A_at_scan_points(A_at_scan_points)
    groups = metric_subgroups(cs, max_delta=5)
    for group in groups.result_groups:
        best_subsym = group["best_subsym"]
        cb_op = group["cb_op_inp_best"]
        expts = ExperimentList([Experiment(crystal=cryst)])
        reindexed_expts = reindex_experiments(
            experiments=expts, cb_op=cb_op, space_group=best_subsym.space_group()
        )
        assert (
            reindexed_expts[0]
            .crystal.get_crystal_symmetry()
            .is_similar_symmetry(best_subsym)
        )
        # Check that the scan-varying A matrices have been copied as well
        assert cryst.num_scan_points == n_scan_points


def test_reindex_cb_op_exit(dials_data):
    data_dir = dials_data("insulin_processed", pathlib=True)

    # Want a SystemExit, rather than an uncaught exception
    with pytest.raises(SystemExit) as e:
        dials.command_line.reindex.run(
            [
                "change_of_basis_op=a+2*b+c,-b+c,a",
                str(data_dir / "scaled.expt"),
                str(data_dir / "scaled.refl"),
            ]
        )
    # Make sure this is the expected error rather than e.g. an argument error
    assert "Unsuitable value" in str(e.value)


def test_reindex_reference_multi_crystal(dials_data, tmp_path):
    mcp = dials_data("multi_crystal_proteinase_k", pathlib=True)
    args = ["dials.cosym", "space_group=P4"]
    for i in [1, 2, 3, 4]:
        args.append(mcp / f"experiments_{i}.json")
        args.append(mcp / f"reflections_{i}.pickle")
    result = procrunner.run(args, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("symmetrized.refl").is_file()
    assert tmp_path.joinpath("symmetrized.expt").is_file()

    args1 = [
        "dials.cosym",
        "space_group=P4",
        "output.experiments=sub_1.expt",
        "output.reflections=sub_1.refl",
    ]
    for i in [5, 7, 8, 10]:
        args1.append(mcp / f"experiments_{i}.json")
        args1.append(mcp / f"reflections_{i}.pickle")
    result1 = procrunner.run(args1, working_directory=tmp_path)
    assert not result1.returncode and not result1.stderr
    assert tmp_path.joinpath("sub_1.expt").is_file()
    assert tmp_path.joinpath("sub_1.refl").is_file()

    args2 = [
        "dials.reindex",
        "symmetrized.refl",
        "symmetrized.expt",
        "reference.experiments=sub_1.expt",
        "reference.reflections=sub_1.refl",
    ]
    result2 = procrunner.run(args2, working_directory=tmp_path)
    assert not result2.returncode and not result2.stderr
    assert tmp_path.joinpath("reindexed.expt").is_file()
    assert tmp_path.joinpath("reindexed.refl").is_file()



from __future__ import annotations

from typing import List

import libtbx.phil
from dxtbx.model import ExperimentList

import dials.util
from dials.command_line.dials_import import ManualGeometryUpdater
from dials.util.options import ArgumentParser, flatten_experiments

help_message = """
"""

phil_scope = libtbx.phil.parse(
    """
include scope dials.util.options.geometry_phil_scope
output {
  experiments = modified.expt
    .type = path
}
""",
    process_includes=True,
)


def update(
    experiments: ExperimentList, new_params: libtbx.phil.scope_extract
) -> ExperimentList:

    """
    Modify detector, beam, goniometer and scan in experiments with the values in new_params
    """

    update_geometry = ManualGeometryUpdater(new_params)

    imagesets = experiments.imagesets()

    for imageset in imagesets:
        imageset_new = update_geometry(imageset)
        imageset.set_detector(imageset_new.get_detector())
        imageset.set_beam(imageset_new.get_beam())
        imageset.set_goniometer(imageset_new.get_goniometer())
        imageset.set_scan(imageset_new.get_scan())

    return experiments


@dials.util.show_mail_handle_errors()
def run(args: List[str] = None, phil: libtbx.phil.scope = phil_scope) -> None:

    usage = "dials.modify_geometry [options] models.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, _ = parser.parse_args(args, show_diff_phil=True)

    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        exit(0)

    new_experiments = update(experiments, params)

    if len(new_experiments):
        print(f"Saving modified experiments to {params.output.experiments}")
        new_experiments.as_file(params.output.experiments)


if __name__ == "__main__":
    run()



"""Test for new experiment identifier features"""
from __future__ import annotations

from dxtbx.model import Experiment, ExperimentList

from dials.array_family import flex


def test_selection_identifier_propagation():
    """
    If experiment idenfitiers are set, then when selecting on a reflection
    table, copy across any for the same id.

    If no identifiers are set, then nothing should happen."""

    # Test when identifiers & id column set.
    refl = flex.reflection_table()
    refl["id"] = flex.int([0, 0, 0, 1, 1, 1, 2, 2, 2])
    refl.experiment_identifiers()[0] = "0"
    refl.experiment_identifiers()[1] = "1"
    refl.experiment_identifiers()[2] = "2"

    new_refl = refl.select(refl["id"] == 1)
    assert list(new_refl.experiment_identifiers().keys()) == [1]
    assert list(new_refl.experiment_identifiers().values()) == ["1"]

    # test with no identifiers set, but there is an id column
    refl = flex.reflection_table()
    refl["id"] = flex.int([0, 0, 0, 1, 1, 1, 2, 2, 2])
    new_refl = refl.select(refl["id"] == 0)
    assert new_refl.ncols() == 1
    assert list(new_refl.experiment_identifiers().keys()) == []
    assert list(new_refl.experiment_identifiers().values()) == []

    # Test selections still work when nothing id related is set.
    refl = flex.reflection_table()
    refl["x"] = flex.int([0, 0, 0, 1, 1, 1, 2, 2, 2])
    new_refl = refl.select(refl["x"] == 0)
    assert new_refl.ncols() == 1
    assert list(new_refl.experiment_identifiers().keys()) == []
    assert list(new_refl.experiment_identifiers().values()) == []


def test_select_using_experiment():
    """
    want to be able to do:
    for exp in experiments:
        refl = reflections.select(exp)

    rather than:
    for i, _ in enumerate(experiments):
        refl = reflections.select(refl["id"] == i)
    """
    exp1 = Experiment(identifier="1")
    exp2 = Experiment(identifier="2")

    refl = flex.reflection_table()
    refl["id"] = flex.int([0, 0, 0, 1, 1, 1])
    refl["idx"] = flex.int([0, 1, 2, 3, 4, 5])
    refl.experiment_identifiers()[0] = "0"
    refl.experiment_identifiers()[1] = "1"

    sel_refl = refl.select(exp1)
    assert list(sel_refl.experiment_identifiers().keys()) == [1]
    assert list(sel_refl.experiment_identifiers().values()) == ["1"]
    assert list(sel_refl["idx"]) == [3, 4, 5]

    # Test returns nothing if no match
    sel_refl = refl.select(exp2)
    assert list(sel_refl.experiment_identifiers().keys()) == []
    assert list(sel_refl.experiment_identifiers().values()) == []
    assert not sel_refl

    # Test returns nothing if no identifiers set. Thought is that this is
    # more useful than raising a dxtbx assert error, as it allows handling within
    # python in case identifiers not set.
    refl = flex.reflection_table()
    refl["idx"] = flex.int([0, 1, 2, 3, 4, 5])
    sel_refl = refl.select(exp1)
    assert list(sel_refl.experiment_identifiers().keys()) == []
    assert list(sel_refl.experiment_identifiers().values()) == []
    assert not sel_refl


def test_select_using_experiments():
    """Test selecting using an experimentlist

    e.g refl = reflections.select(experiments) to select a subset of the
    data from an experimentlist."""

    refl = flex.reflection_table()
    refl["id"] = flex.int([0, 0, 0, 1, 1, 1, 2, 2, 2])
    refl["idx"] = flex.int([0, 1, 2, 3, 4, 5, 6, 7, 8])
    refl.experiment_identifiers()[0] = "0"
    refl.experiment_identifiers()[1] = "1"
    refl.experiment_identifiers()[2] = "2"
    experiments = ExperimentList(
        [
            Experiment(identifier="0"),
            Experiment(identifier="2"),
            Experiment(identifier="3"),
        ]
    )

    sel_refl = refl.select(experiments)
    assert list(sel_refl.experiment_identifiers().keys()) == [0, 2]
    assert list(sel_refl.experiment_identifiers().values()) == ["0", "2"]
    assert list(sel_refl["idx"]) == [0, 1, 2, 6, 7, 8]



#
# See https://github.com/dials/dials/wiki/pytest for documentation on how to
# write and run pytest tests, and an overview of the available features.
#


from __future__ import annotations

import multiprocessing
import os
import sys
import warnings
from pathlib import Path

import pytest
from _pytest.outcomes import Skipped

# https://stackoverflow.com/a/40846742
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

if sys.version_info[:2] == (3, 7) and sys.platform == "darwin":
    multiprocessing.set_start_method("forkserver")

collect_ignore = []


def pytest_configure(config):
    if not config.pluginmanager.hasplugin("dials_data"):

        @pytest.fixture(scope="session")
        def dials_data():
            pytest.skip("This test requires the dials_data package to be installed")

        globals()["dials_data"] = dials_data
    config.addinivalue_line(
        "markers", "xfel: Mark test to run xfail if xfel module is missing"
    )


def pytest_collection_modifyitems(config, items):
    # Attempt to import xfel
    try:
        import xfel  # noqa: F401
    except (Skipped, ModuleNotFoundError):
        # We don't have XFEL
        xfail_marker = pytest.mark.xfail(reason="XFEL module not present")
        for item in items:
            if item.get_closest_marker("xfel"):
                item.add_marker(xfail_marker)


@pytest.fixture(scope="session")
def dials_regression():
    """Return the absolute path to the dials_regression module as a string.
    Skip the test if dials_regression is not installed."""

    if "DIALS_REGRESSION" in os.environ:
        return os.environ["DIALS_REGRESSION"]

    try:
        import dials_regression as dr

        return os.path.dirname(dr.__file__)
    except ImportError:
        pass  # dials_regression not configured
    try:
        import socket

        reference_copy = "/dls/science/groups/scisoft/DIALS/repositories/git-reference/dials_regression"
        if (
            os.name == "posix"
            and socket.gethostname().endswith(".diamond.ac.uk")
            and os.path.exists(reference_copy)
        ):
            return reference_copy
    except ImportError:
        pass  # Cannot tell whether in DLS network or not
    pytest.skip("dials_regression required for this test")


@pytest.fixture
def run_in_tmp_path(tmp_path) -> Path:
    """
    A fixture to change the working directory for the test to a temporary directory.

    The original working directory is restored upon teardown of the fixture.

    Args:
        tmp_path: Pytest tmp_path fixture, see
                  https://docs.pytest.org/en/latest/how-to/tmp_path.html

    Yields:
        The path to the temporary working directory defined by tmp_path.
    """
    cwd = Path.cwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(cwd)



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1
# DIALS_ENABLE_COMMAND_LINE_COMPLETION
# LIBTBX_SET_DISPATCHER_NAME dials.reciprocal_lattice_viewer
# LIBTBX_SET_DISPATCHER_NAME dials.rlv

from __future__ import annotations

import copy
import os

import wxtbx.app
from scitbx.array_family import flex

import dials.util.log
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.reciprocal_lattice.viewer import ReciprocalLatticeViewer, phil_scope

help_message = """
Visualise the strong spots from spotfinding in reciprocal space.

Examples::

  dials.reciprocal_lattice_viewer imported.expt strong.refl

  dials.reciprocal_lattice_viewer indexed.expt indexed.refl
"""


@dials.util.show_mail_handle_errors()
def run(args=None):
    dials.util.log.print_banner()
    usage = "dials.reciprocal_lattice_viewer [options] models.expt observations.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(experiments) == 0 or len(reflections) == 0:
        parser.print_help()
        exit(0)

    if len(reflections) > 1:
        assert len(reflections) == len(experiments)
        for i in range(len(reflections)):
            reflections[i]["imageset_id"] = flex.int(len(reflections[i]), i)
            if i > 0:
                reflections[0].extend(reflections[i])
    elif "imageset_id" not in reflections[0]:
        reflections[0]["imageset_id"] = reflections[0]["id"]

    reflections = reflections[0]

    a = wxtbx.app.CCTBXApp(0)
    a.settings = params
    f = ReciprocalLatticeViewer(
        None,
        -1,
        os.path.realpath(params.input.reflections[0].filename),
        size=(1024, 768),
        settings=copy.deepcopy(params),
    )
    f.load_models(experiments, reflections)
    f.Show()
    a.SetTopWindow(f)
    a.MainLoop()


if __name__ == "__main__":
    run()



from __future__ import annotations

import pickle


def reflections(obj, outfile):
    """
    Dump the given object to file

    :param obj: The reflection list to dump
    :param outfile: The output file name or file object
    """
    if isinstance(outfile, str):
        with open(outfile, "wb") as outfile:
            pickle.dump(obj, outfile, pickle.HIGHEST_PROTOCOL)

    # Otherwise assume the input is a file and write to it
    else:
        pickle.dump(obj, outfile, pickle.HIGHEST_PROTOCOL)


def reference(obj, outfile):
    """
    Dump the given object to file

    :param obj: The reference list to dump
    :param outfile: The output file name or file object
    """

    if isinstance(outfile, str):
        with open(outfile, "wb") as outfile:
            pickle.dump(obj, outfile, pickle.HIGHEST_PROTOCOL)

    # Otherwise assume the input is a file and write to it
    else:
        pickle.dump(obj, outfile, pickle.HIGHEST_PROTOCOL)



from __future__ import annotations

import iotbx.phil

import dials.util
from dials.util import Sorry

master_phil_scope = iotbx.phil.parse(
    """
scan_range = None
  .help = "The range of images to use in indexing. Number of arguments"
          "must be a factor of two. Specifying \"0 0\" will use all images"
          "by default. The given range follows C conventions"
          "(e.g. j0 <= j < j1)."
  .type = ints(size=2)
  .multiple = True
first_n_reflections = None
  .type = int(value_min=0)
crystal_id = None
  .type = int
output {
  file_name = centroids.png
    .type = path
  show_plot = False
    .type = bool
  dpi = 300
    .type = int(value_min=0)
  size_inches = 10,10
    .type = floats(size=2, value_min=0)
  marker_size = 10
    .type = float(value_min=0)
}
"""
)


@dials.util.show_mail_handle_errors()
def run(args=None):
    usage = "dials.plot_reflections models.expt observations.refl [options]"
    from scitbx import matrix
    from scitbx.array_family import flex

    from dials.util.options import (
        ArgumentParser,
        reflections_and_experiments_from_files,
    )

    parser = ArgumentParser(
        usage=usage,
        phil=master_phil_scope,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )
    if len(experiments.imagesets()) > 0:
        imageset = experiments.imagesets()[0]
        imageset.set_detector(experiments[0].detector)
        imageset.set_beam(experiments[0].beam)
        imageset.set_goniometer(experiments[0].goniometer)
    else:
        parser.print_help()
        return

    detector = imageset.get_detector()
    scan = imageset.get_scan()

    panel_origin_shifts = {0: (0, 0, 0)}
    try:
        hierarchy = detector.hierarchy()
    except AttributeError:
        hierarchy = None
    for i_panel in range(1, len(detector)):
        origin_shift = matrix.col(detector[0].get_origin()) - matrix.col(
            detector[i_panel].get_origin()
        )
        panel_origin_shifts[i_panel] = origin_shift

    observed_xyz = flex.vec3_double()
    predicted_xyz = flex.vec3_double()

    for reflection_list in reflections:

        if len(params.scan_range):
            sel = flex.bool(len(reflection_list), False)

            xyzcal_px = None
            xyzobs_px = None

            if "xyzcal.px" in reflection_list:
                xyzcal_px = reflection_list["xyzcal.px"]
            if "xyzobs.px.value" in reflection_list:
                xyzobs_px = reflection_list["xyzobs.px.value"]

            if xyzcal_px is not None and not xyzcal_px.norms().all_eq(0):
                centroids_frame = xyzcal_px.parts()[2]
            elif xyzobs_px is not None and not xyzobs_px.norms().all_eq(0):
                centroids_frame = xyzobs_px.parts()[2]
            else:
                raise Sorry("No pixel coordinates given in input reflections.")

            for scan_range in params.scan_range:
                if scan_range is None:
                    continue
                range_start, range_end = scan_range
                sel |= (centroids_frame >= range_start) & (centroids_frame < range_end)
            reflection_list = reflection_list.select(sel)
        if params.first_n_reflections is not None:
            centroid_positions = reflection_list.centroid_position()
            centroids_frame = centroid_positions.parts()[2]
            perm = flex.sort_permutation(centroids_frame)
            perm = perm[: min(reflection_list.size(), params.first_n_reflections)]
            reflection_list = reflection_list.select(perm)
        if params.crystal_id is not None:
            reflection_list = reflection_list.select(
                reflection_list["id"] == params.crystal_id
            )

        xyzobs_px = None
        xyzobs_mm = None
        xyzcal_mm = None

        if "xyzobs.px.value" in reflection_list:
            xyzobs_px = reflection_list["xyzobs.px.value"]
        if "xyzcal.mm" in reflection_list:
            xyzcal_mm = reflection_list["xyzcal.mm"]
        if "xyzobs.mm.value" in reflection_list:
            xyzobs_mm = reflection_list["xyzobs.mm.value"]

        panel_ids = reflection_list["panel"]
        if xyzobs_mm is None and xyzobs_px is not None:
            xyzobs_mm = flex.vec3_double()
            for i_panel in range(len(detector)):
                xyzobs_px_panel = xyzobs_px.select(panel_ids == i_panel)

                from dials.algorithms.centroid import centroid_px_to_mm_panel

                xyzobs_mm_panel, _, _ = centroid_px_to_mm_panel(
                    detector[i_panel],
                    scan,
                    xyzobs_px_panel,
                    flex.vec3_double(xyzobs_px_panel.size()),
                    flex.vec3_double(xyzobs_px_panel.size()),
                )
                xyzobs_mm.extend(xyzobs_mm_panel)

        if xyzobs_mm is not None:
            observed_xyz.extend(xyzobs_mm)
        if xyzcal_mm is not None:
            predicted_xyz.extend(xyzcal_mm)

    obs_x, obs_y, _ = observed_xyz.parts()
    pred_x, pred_y, _ = predicted_xyz.parts()

    try:
        import matplotlib

        if not params.output.show_plot:
            # http://matplotlib.org/faq/howto_faq.html#generate-images-without-having-a-window-appear
            matplotlib.use("Agg")  # use a non-interactive backend
        import matplotlib.pyplot as plt
    except ImportError:
        raise Sorry("matplotlib must be installed to generate a plot.")

    fig = plt.figure()
    fig.set_size_inches(params.output.size_inches)
    fig.set_dpi(params.output.dpi)
    plt.axes().set_aspect("equal")
    marker_size = params.output.marker_size
    if obs_x.size():
        plt.scatter(
            obs_x,
            obs_y,
            marker="o",
            edgecolors="black",
            c="white",
            s=marker_size,
            alpha=1,
        )
    if pred_x.size():
        plt.scatter(pred_x, pred_y, marker="+", s=marker_size, c="blue")
    xmax = max(
        [
            detector[i_panel].get_image_size_mm()[0] + panel_origin_shifts[i_panel][0]
            for i_panel in range(len(detector))
        ]
    )
    ymax = max(
        [
            detector[i_panel].get_image_size_mm()[1] + panel_origin_shifts[i_panel][1]
            for i_panel in range(len(detector))
        ]
    )
    try:
        beam_centre = hierarchy.get_beam_centre(imageset.get_beam().get_s0())
    except Exception:
        beam_centre = detector[0].get_beam_centre(imageset.get_beam().get_s0())
    plt.scatter([beam_centre[0]], [beam_centre[1]], marker="+", c="blue", s=100)
    plt.xlim(0, xmax)
    plt.ylim(0, ymax)
    plt.gca().invert_yaxis()
    plt.title("Centroid x,y-coordinates")
    plt.xlabel("x-coordinate (mm)")
    plt.ylabel("y-coordinate (mm)")
    if params.output.file_name is not None:
        plt.savefig(
            params.output.file_name,
            dpi=params.output.dpi,
            bbox_inches="tight",
        )
    if params.output.show_plot:
        plt.show()


if __name__ == "__main__":
    run()



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1
# DIALS_ENABLE_COMMAND_LINE_COMPLETION

from __future__ import annotations

import copy
import sys

import wx

import gltbx
import gltbx.gl as gl
import libtbx.phil
import wxtbx.utils
from dxtbx.model import MultiAxisGoniometer
from scitbx.array_family import flex
from scitbx.math import minimum_covering_sphere

from dials.util import show_mail_handle_errors, wx_viewer

help_message = """
"""

phil_scope = libtbx.phil.parse(
    """
  angle = None
    .type = float
    .multiple = True
  detector_distance = None
    .type = float
  z_min = None
    .type = float
  z_max = None
    .type = float
  marker_size = 100
    .type = int(value_min=1)
  autospin = False
    .type = bool
  predict = False
    .type = bool
  prediction_width = None
    .type = float(value_min=0)
    .help = "Width of prediction window (degrees)"
  show_panel_axes = False
    .type = bool
    .help = "Plot the fast, slow and normal vectors for each panel."
  show_crystal_axes = True
    .type = bool
    .help = "Plot the crystal reciprocal cell axes"
  require_images = True
    .type = bool
    .help = "Flag which can be set to False to launch image viewer without
      checking the image format (needed for some image format classes).
      Alternative to DIALS_EXPORT_DO_NOT_CHECK_FORMAT environment variable."
"""
)


def settings():
    return phil_scope.fetch().extract()


class render_3d:
    def __init__(self):
        self.goniometer_orig = None
        self.crystal = None

    def load_imageset(self, imageset, crystal=None):
        self.imageset = imageset
        if crystal is not None:
            self.crystal = crystal
        if self.imageset.get_goniometer() is not None:
            from scitbx import matrix

            gonio = self.imageset.get_goniometer()
            axis = matrix.col(gonio.get_rotation_axis()).elems
            self.viewer.set_rotation_axis(axis)
        self.viewer.set_beam_vector(self.imageset.get_beam().get_s0())

        self.set_points()

    def set_goniometer_points(self):
        gonio = self.imageset.get_goniometer()
        detector = self.imageset.get_detector()
        beam = self.imageset.get_beam()

        angle = self.settings.angle
        if angle:
            assert len(angle) == len(gonio.get_angles())
            gonio.set_angles(angle)

        distance = self.settings.detector_distance
        if distance:
            import math

            from scitbx import matrix

            p_id = detector.get_panel_intersection(beam.get_s0())
            if p_id >= 0:
                if len(detector) > 1:
                    p = detector.hierarchy()
                else:
                    p = detector[0]
                d_normal = matrix.col(p.get_normal())
                d_origin = matrix.col(p.get_origin())
                d_distance = math.fabs(
                    d_origin.dot(d_normal) - p.get_directed_distance()
                )
                assert d_distance < 0.001, d_distance
                translation = d_normal * (distance - p.get_directed_distance())
                new_origin = d_origin + translation
                d_distance = math.fabs(new_origin.dot(d_normal) - distance)
                assert d_distance < 0.001, d_distance
                p.set_frame(p.get_fast_axis(), p.get_slow_axis(), new_origin.elems)

        gonio_masker = (
            self.imageset.get_format_class()
            .get_instance(self.imageset.paths()[0], **self.imageset.data().get_params())
            .get_masker(goniometer=gonio)
        )
        if gonio_masker is None:
            return

        points = gonio_masker.extrema_at_scan_angle(
            gonio.get_angles()[gonio.get_scan_axis()]
        )
        points.insert(0, (0, 0, 0))

        line_i_seqs = flex.vec2_double((0, i) for i in range(1, points.size()))
        line_i_seqs += (self.viewer.points.size(), self.viewer.points.size())
        for i_seqs in line_i_seqs:
            self.viewer.line_i_seqs.append([int(i_seq) for i_seq in i_seqs])
            self.viewer.line_colors[(i_seqs)] = (100 / 255, 120 / 255, 255 / 255)

        self.viewer.points.extend(points)

        shadow = gonio_masker.project_extrema(
            detector, gonio.get_angles()[gonio.get_scan_axis()]
        )

        for shadow_points, p in zip(shadow, detector):
            n = self.viewer.points.size()
            line_i_seqs = []
            line_colors = {}
            for i in range(shadow_points.size()):
                if i < shadow_points.size() - 1:
                    line_i_seqs.append((n + i, n + i + 1))
                else:
                    line_i_seqs.append((n + i, n))
                line_colors[line_i_seqs[-1]] = (1, 1, 1)
            self.viewer.line_colors.update(line_colors)
            self.viewer.line_i_seqs.extend(line_i_seqs)
            self.viewer.points.extend(
                p.get_lab_coord(shadow_points * p.get_pixel_size()[0])
            )

    def set_detector_points(self):
        detector = self.imageset.get_detector()
        points = flex.vec3_double()
        line_i_seqs = flex.vec2_double()
        i = 0
        for p in detector:
            image_size = p.get_image_size_mm()
            points.append(p.get_lab_coord((0, 0)))
            points.append(p.get_lab_coord((0, image_size[1])))
            points.append(p.get_lab_coord(image_size))
            points.append(p.get_lab_coord((image_size[0], 0)))
            line_i_seqs.append((i, i + 1))
            line_i_seqs.append((i + 1, i + 2))
            line_i_seqs.append((i + 2, i + 3))
            line_i_seqs.append((i + 3, i))
            i += 4
        line_i_seqs += (self.viewer.points.size(), self.viewer.points.size())
        self.viewer.points.extend(points)
        for i_seqs in line_i_seqs:
            self.viewer.line_i_seqs.append([int(i_seq) for i_seq in i_seqs])

    def set_points(self):
        # reset point/line lists
        self.viewer.reset()

        try:
            self.set_goniometer_points()
        except Exception:
            pass
        self.set_detector_points()
        self.viewer.update_minimum_covering_sphere()
        if self.settings.predict and self.crystal is not None:
            self.set_reflection_points()

    def set_reflection_points(self):
        import time

        t0 = time.time()
        predicted = self.predict()
        if predicted is None:
            return
        t1 = time.time()
        print("Predicted %i reflections in %.2f s" % (predicted.size(), (t1 - t0)))
        xc, yc, zc = predicted["xyzcal.mm"].parts()
        xycal = flex.vec2_double(xc, yc)
        panel = predicted["panel"]
        detector = self.imageset.get_detector()
        for pid, p in enumerate(detector):
            self.viewer.points.extend(p.get_lab_coord(xycal.select(panel == pid)))

    def predict(self):
        assert self.crystal is not None
        from dxtbx.model.experiment_list import Experiment

        imageset = self.imageset
        scan = copy.deepcopy(imageset.get_scan())
        gonio = imageset.get_goniometer()
        prediction_width = self.settings.prediction_width
        if prediction_width is None:
            prediction_width = scan.get_oscillation()[1]
        if isinstance(gonio, MultiAxisGoniometer):
            scan_angle = gonio.get_angles()[gonio.get_scan_axis()]
        else:
            return
        scan.set_oscillation((scan_angle, prediction_width))
        expt = Experiment(
            imageset=imageset,
            crystal=self.crystal,
            detector=imageset.get_detector(),
            beam=imageset.get_beam(),
            scan=scan[:1],
            goniometer=imageset.get_goniometer(),
        )

        # Populate the reflection table with predictions
        from dials.array_family import flex

        predicted = flex.reflection_table.from_predictions(expt, force_static=True)
        predicted["id"] = flex.int(len(predicted), 0)
        return predicted


class ExperimentViewer(wx.Frame, render_3d):
    def __init__(self, *args, **kwds):
        wx.Frame.__init__(self, *args, **kwds)
        render_3d.__init__(self)
        self.parent = self.GetParent()
        self.statusbar = self.CreateStatusBar()
        self.sizer = wx.BoxSizer(wx.HORIZONTAL)

        app = wx.GetApp()
        if getattr(app, "settings", None) is not None:
            # XXX copying the initial settings avoids awkward interactions when
            # multiple viewer windows are opened
            self.settings = copy.deepcopy(app.settings)
        else:
            self.settings = settings()

        self.create_settings_panel()
        self.sizer.Add(self.settings_panel, 0, wx.EXPAND)
        self.create_viewer_panel()
        self.sizer.Add(self.viewer, 1, wx.EXPAND | wx.ALL)
        self.SetSizer(self.sizer)
        self.sizer.SetSizeHints(self)
        self.Bind(wx.EVT_CLOSE, self.OnClose, self)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy, self)
        self.Bind(wx.EVT_ACTIVATE, self.OnActive)
        self.viewer.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.viewer.SetFocus()

    def load_imageset(self, imageset, crystal=None):
        render_3d.load_imageset(self, imageset, crystal=crystal)
        self.settings_panel.add_goniometer_controls(imageset.get_goniometer())

    def OnActive(self, event):
        if self.IsShown() and type(self.viewer).__name__ != "_wxPyDeadObject":
            self.viewer.Refresh()

    def OnClose(self, event):
        self.Unbind(wx.EVT_ACTIVATE)
        self.Destroy()
        event.Skip()

    def OnDestroy(self, event):
        if self.parent is not None:
            self.parent.viewer = None
        event.Skip()

    def OnKeyDown(self, event):
        key = event.GetUnicodeKey()
        if key == wx.WXK_NONE:
            key = event.GetKeyCode()
        dxs = {wx.WXK_LEFT: -1, wx.WXK_RIGHT: +1, wx.WXK_UP: 0, wx.WXK_DOWN: 0}
        dys = {wx.WXK_LEFT: 0, wx.WXK_RIGHT: 0, wx.WXK_UP: +1, wx.WXK_DOWN: -1}

        if key in dxs:
            dx = dxs[key]
            dy = dys[key]
            if event.ShiftDown():
                scale = 0.1
            else:
                scale = 1.0
            self.do_Step(dx, dy, scale)

    def do_Step(self, dx, dy, scale):
        v = self.viewer
        rc = v.rotation_center
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gltbx.util.rotate_object_about_eye_x_and_y(
            scale, rc[0], rc[1], rc[2], dx, dy, 0, 0
        )
        v.OnRedraw()

    def create_viewer_panel(self):
        self.viewer = GeometryWindow(
            settings=self.settings,
            parent=self,
            size=(800, 600),
            # orthographic=True
        )

    def create_settings_panel(self):
        self.settings_panel = settings_window(self, -1, style=wx.RAISED_BORDER)

    def set_points(self):
        render_3d.set_points(self)

    def update_settings(self, *args, **kwds):
        self.set_points()
        self.viewer.update_settings(*args, **kwds)


class settings_window(wxtbx.utils.SettingsPanel):
    def __init__(self, *args, **kwds):
        wxtbx.utils.SettingsPanel.__init__(self, *args, **kwds)
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def OnChar(self, event):
        self.GetParent().viewer.OnChar(event)

    def add_controls(self):

        ctrls = self.create_controls(setting="show_panel_axes", label="Show panel axes")
        self.panel_sizer.Add(ctrls[0], 0, wx.ALL, 5)

        ctrls = self.create_controls(
            setting="show_crystal_axes", label="Show crystal axes"
        )
        self.panel_sizer.Add(ctrls[0], 0, wx.ALL, 5)

    def add_goniometer_controls(self, goniometer):
        from wx.lib.agw import floatspin

        self.distance_ctrl = floatspin.FloatSpin(parent=self, increment=1, digits=2)
        self.distance_ctrl.SetValue(self.settings.detector_distance)
        self.distance_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.distance_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Detector distance")
        box.Add(self.distance_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.distance_ctrl)

        self.rotation_angle_ctrls = []
        if isinstance(goniometer, MultiAxisGoniometer):
            names = goniometer.get_names()
            axes = goniometer.get_axes()
            angles = goniometer.get_angles()
            for name, axis, angle in zip(names, axes, angles):
                ctrl = floatspin.FloatSpin(parent=self, increment=1, digits=3)
                ctrl.SetValue(angle)
                ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
                if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
                    ctrl.SetBackgroundColour(self.GetBackgroundColour())
                box = wx.BoxSizer(wx.HORIZONTAL)
                self.panel_sizer.Add(box)
                label = wx.StaticText(self, -1, f"{name} angle")
                box.Add(ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
                box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
                self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, ctrl)
                self.rotation_angle_ctrls.append(ctrl)

    def OnChangeSettings(self, event):
        self.settings.detector_distance = self.distance_ctrl.GetValue()
        for i, ctrl in enumerate(self.rotation_angle_ctrls):
            self.settings.angle[i] = ctrl.GetValue()
        self.parent.update_settings()


class GeometryWindow(wx_viewer.show_points_and_lines_mixin):
    def __init__(self, settings, *args, **kwds):
        super().__init__(*args, **kwds)
        self.settings = settings
        self.points = flex.vec3_double()
        self.colors = None
        self.rotation_axis = None
        self.beam_vector = None
        self.flag_show_minimum_covering_sphere = False
        self.minimum_covering_sphere = None
        self.field_of_view_y = 20
        if self.settings.autospin:
            self.autospin_allowed = True
            self.yspin = 1
            self.xspin = 1
            self.autospin = True

    def reset(self):
        self.labels = []
        self.points = flex.vec3_double()
        self.line_i_seqs = []
        self.line_colors = {}
        self.spheres = []
        self.labels_display_list = None
        self.points_display_list = None
        self.lines_display_list = None

    def set_points(self, points):
        self.points = points
        self.points_display_list = None
        if self.minimum_covering_sphere is None:
            self.update_minimum_covering_sphere()

    def set_colors(self, colors):
        assert len(colors) == len(self.points)
        self.colors = colors

    def draw_points(self):
        if self.points_display_list is None:
            self.points_display_list = gltbx.gl_managed.display_list()
            self.points_display_list.compile()
            gl.glLineWidth(1)
            if self.colors is None:
                self.colors = flex.vec3_double(len(self.points), (1, 1, 1))
            for point, color in zip(self.points, self.colors):
                self.draw_cross_at(point, color=color)
            self.points_display_list.end()
        self.points_display_list.call()

    def set_rotation_axis(self, axis):
        self.rotation_axis = axis

    def set_beam_vector(self, beam):
        self.beam_vector = beam

    # --- user input and settings
    def update_settings(self):
        self.points_display_list = None
        self.Refresh()

    def update_minimum_covering_sphere(self):
        self.minimum_covering_sphere = minimum_covering_sphere(
            self.points, epsilon=1e-3
        )

    def draw_cross_at(self, xyz, color=(1, 1, 1), f=None):
        (x, y, z) = xyz
        if f is None:
            f = 0.01 * self.settings.marker_size
        wx_viewer.show_points_and_lines_mixin.draw_cross_at(
            self, (x, y, z), color=color, f=f
        )

    def DrawGL(self):
        wx_viewer.show_points_and_lines_mixin.DrawGL(self)
        gonio = self.parent.imageset.get_goniometer()
        beam = self.parent.imageset.get_beam()
        detector = self.parent.imageset.get_detector()
        if self.settings.show_panel_axes and detector is not None:
            for p in detector:
                image_size = p.get_image_size_mm()
                from scitbx import matrix

                o = matrix.col(p.get_lab_coord((0, 0)))
                f = matrix.col(p.get_lab_coord((image_size[0], 0)))
                _f = (f - o).normalize()
                s = matrix.col(p.get_lab_coord((0, image_size[1])))
                _s = (s - o).normalize()
                _n = _f.cross(_s)
                n = _n * (0.25 * (f.length() + s.length())) + o
                self.draw_lab_axis(o.elems, f.elems, "FAST")
                self.draw_lab_axis(o.elems, s.elems, "SLOW")
                self.draw_lab_axis(o.elems, n.elems, "NORM")

        from scitbx import matrix

        if isinstance(gonio, MultiAxisGoniometer):
            R = matrix.identity(3)
            names = reversed(gonio.get_names())
            axes = reversed(gonio.get_axes())
            angles = reversed(gonio.get_angles())
            for name, axis, angle in zip(names, axes, angles):
                axis = R * matrix.col(axis)
                self.draw_axis(axis.elems, name)
                R = axis.axis_and_angle_as_r3_rotation_matrix(angle, deg=True) * R
        elif gonio is not None:
            axis = matrix.col(gonio.get_rotation_axis())
            self.draw_axis(axis.elems, "phi")
        self.draw_axis(beam.get_s0(), "beam")
        crystal = self.parent.crystal
        if self.settings.show_crystal_axes and crystal is not None:
            crystal = copy.deepcopy(crystal)
            scan = self.parent.imageset.get_scan()
            fixed_rotation = matrix.sqr(gonio.get_fixed_rotation())
            setting_rotation = matrix.sqr(gonio.get_setting_rotation())
            rotation_axis = matrix.col(gonio.get_rotation_axis_datum())
            if isinstance(gonio, MultiAxisGoniometer):
                angle = gonio.get_angles()[gonio.get_scan_axis()]
            else:
                angle = scan.get_oscillation()[0]
            rotation_matrix = rotation_axis.axis_and_angle_as_r3_rotation_matrix(
                angle, deg=True
            )
            U0 = matrix.sqr(crystal.get_U())

            # Goniometer datum setting [D] at which the orientation was determined
            D = (setting_rotation * rotation_matrix * fixed_rotation).inverse()

            U = D.inverse() * U0
            B = matrix.sqr(crystal.get_B())
            a_star = U * B * matrix.col((1, 0, 0))
            b_star = U * B * matrix.col((0, 1, 0))
            c_star = U * B * matrix.col((0, 0, 1))
            color = (1.0, 0.0, 0.0)  # red
            self.draw_axis(a_star.normalize() * 0.5, "a*", color=color)
            self.draw_axis(b_star.normalize() * 0.5, "b*", color=color)
            self.draw_axis(c_star.normalize() * 0.5, "c*", color=color)

    def draw_axis(self, axis, label, color=(1.0, 1.0, 1.0)):
        if self.minimum_covering_sphere is None:
            self.update_minimum_covering_sphere()
        s = self.minimum_covering_sphere
        scale = max(max(s.box_max()), abs(min(s.box_min())))
        gltbx.fonts.ucs_bitmap_8x13.setup_call_lists()
        gl.glDisable(gl.GL_LIGHTING)
        gl.glColor3f(*color)
        gl.glLineWidth(1.0)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(0.0, 0.0, 0.0)
        gl.glVertex3f(axis[0] * scale, axis[1] * scale, axis[2] * scale)
        gl.glEnd()
        gl.glRasterPos3f(
            0.5 + axis[0] * scale, 0.2 + axis[1] * scale, 0.2 + axis[2] * scale
        )
        gltbx.fonts.ucs_bitmap_8x13.render_string(label)
        gl.glEnable(gl.GL_LINE_STIPPLE)
        gl.glLineStipple(4, 0xAAAA)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(0.0, 0.0, 0.0)
        gl.glVertex3f(-axis[0] * scale, -axis[1] * scale, -axis[2] * scale)
        gl.glEnd()
        gl.glDisable(gl.GL_LINE_STIPPLE)

    def draw_lab_axis(self, start, end, label):
        mid = tuple([0.5 * (s + e) for s, e in zip(start, end)])
        gltbx.fonts.ucs_bitmap_8x13.setup_call_lists()
        gl.glDisable(gl.GL_LIGHTING)
        gl.glColor3f(1.0, 1.0, 0.0)
        gl.glLineWidth(1.0)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(*start)
        gl.glVertex3f(*end)
        gl.glEnd()
        gl.glRasterPos3f(*mid)
        gltbx.fonts.ucs_bitmap_8x13.render_string(label)

    def rotate_view(self, x1, y1, x2, y2, shift_down=False, scale=0.1):
        super().rotate_view(x1, y1, x2, y2, shift_down=shift_down, scale=scale)

    def OnLeftUp(self, event):
        self.was_dragged = True
        super().OnLeftUp(event)

    def initialize_modelview(self, eye_vector=None, angle=None):
        super().initialize_modelview(eye_vector=eye_vector, angle=angle)
        self.rotation_center = (0, 0, 0)
        self.move_to_center_of_viewport(self.rotation_center)


@show_mail_handle_errors()
def run(args=None):

    import os

    from dials.util.options import ArgumentParser, flatten_experiments

    usage = "dials.geometry_viewer [options] models.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, quick_parse=True, show_diff_phil=True)

    if "DIALS_EXPORT_DO_NOT_CHECK_FORMAT" in os.environ:
        print(
            "\nWarning: use of DIALS_EXPORT_DO_NOT_CHECK_FORMAT environment variable"
            "\nis no longer recommended, the recommended command is require_images=False"
        )
    if not params.require_images or "DIALS_EXPORT_DO_NOT_CHECK_FORMAT" in os.environ:
        check_format = False
    else:
        check_format = True

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=check_format,
        epilog=help_message,
    )

    try:
        params, options = parser.parse_args(show_diff_phil=False)
    except Exception as e:
        print(e)
        print(
            "dials.geometry_viewer help: Error in parsing data may be due to missing \n"
            "files. If so, this may be overcome by require_images=False\n"
        )
        sys.exit()

    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        exit(0)

    imagesets = experiments.imagesets()
    crystal = experiments[0].crystal

    assert len(imagesets) == 1, len(imagesets)
    imageset = imagesets[0]
    gonio = imageset.get_goniometer()
    if not params.detector_distance:
        detector = imageset.get_detector()
        if len(detector) > 1:
            params.detector_distance = detector.hierarchy().get_directed_distance()
        else:
            params.detector_distance = detector[0].get_directed_distance()

    if gonio is not None and not isinstance(gonio, MultiAxisGoniometer):
        from dxtbx.model.goniometer import GoniometerFactory

        gonio = GoniometerFactory.multi_axis(
            axes=flex.vec3_double((gonio.get_rotation_axis(),)),
            angles=flex.double((0,)),
            names=flex.std_string(("GON_OMEGA",)),
            scan_axis=0,
        )
        imageset.set_goniometer(gonio)

    if isinstance(gonio, MultiAxisGoniometer):
        if params.angle:
            assert len(params.angle) == len(gonio.get_angles())
        else:
            for angle in gonio.get_angles():
                params.angle.append(angle)

    import wxtbx.app

    a = wxtbx.app.CCTBXApp(0)
    a.settings = params
    f = ExperimentViewer(None, -1, "Experiment viewer", size=(1024, 768))
    f.load_imageset(imageset, crystal=crystal)
    f.Show()
    a.SetTopWindow(f)
    a.MainLoop()


if __name__ == "__main__":
    run()



from __future__ import annotations

import math

from dxtbx.model.experiment_list import ExperimentListFactory

from dials.algorithms.simulation.reciprocal_space import Simulator


def test_simulation(dials_data):
    experiments = ExperimentListFactory.from_json_file(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json",
        check_format=False,
    )

    sigma_b = 0.058 * math.pi / 180
    sigma_m = 0.157 * math.pi / 180
    n_sigma = 3

    N = 100
    In = 1000
    B = 10
    simulate = Simulator(experiments[0], sigma_b, sigma_m, n_sigma)
    simulate.with_random_intensity(N, In, B, 0, 0, 0)



from __future__ import annotations

import os
from pathlib import Path

import procrunner
import pytest

from dxtbx.model.experiment_list import ExperimentListFactory
from libtbx import phil
from scitbx import matrix

from dials.algorithms.refinement import RefinerFactory
from dials.array_family import flex


def test1(dials_regression, tmp_path):
    data_dir = Path(dials_regression) / "refinement_test_data" / "multi_stills"

    result = procrunner.run(
        [
            "dials.refine",
            data_dir / "combined_experiments.json",
            data_dir / "combined_reflections.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # load results
    reg_exp = ExperimentListFactory.from_json_file(
        data_dir / "regression_experiments.json", check_format=False
    )
    ref_exp = ExperimentListFactory.from_json_file(
        tmp_path / "refined.expt", check_format=False
    )

    # compare results
    tol = 1e-5
    for b1, b2 in zip(reg_exp.beams(), ref_exp.beams()):
        assert b1.is_similar_to(
            b2,
            wavelength_tolerance=tol,
            direction_tolerance=tol,
            polarization_normal_tolerance=tol,
            polarization_fraction_tolerance=tol,
        )
        s0_1 = matrix.col(b1.get_unit_s0())
        s0_2 = matrix.col(b2.get_unit_s0())
        assert s0_1.accute_angle(s0_2, deg=True) < 0.0057  # ~0.1 mrad
    for c1, c2 in zip(reg_exp.crystals(), ref_exp.crystals()):
        assert c1.is_similar_to(c2)

    for d1, d2 in zip(reg_exp.detectors(), ref_exp.detectors()):
        assert d1.is_similar_to(
            d2,
            fast_axis_tolerance=1e-4,
            slow_axis_tolerance=1e-4,
            origin_tolerance=1e-2,
        )


@pytest.mark.skipif(
    os.name == "nt",
    reason="Multiprocessing error on Windows: 'This class cannot be instantiated from Python'",
)
def test_multi_process_refinement_gives_same_results_as_single_process_refinement(
    dials_regression, tmp_path
):
    data_dir = Path(dials_regression) / "refinement_test_data" / "multi_stills"
    cmd = [
        "dials.refine",
        data_dir / "combined_experiments.json",
        data_dir / "combined_reflections.pickle",
        "outlier.algorithm=null",
        "engine=LBFGScurvs",
        "output.reflections=None",
    ]
    result = procrunner.run(
        cmd + ["output.experiments=refined_nproc4.expt", "nproc=4"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    result = procrunner.run(
        cmd + ["output.experiments=refined_nproc1.expt", "nproc=1"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # load results
    nproc1 = ExperimentListFactory.from_json_file(
        tmp_path / "refined_nproc1.expt", check_format=False
    )
    nproc4 = ExperimentListFactory.from_json_file(
        tmp_path / "refined_nproc4.expt", check_format=False
    )

    # compare results
    for b1, b2 in zip(nproc1.beams(), nproc4.beams()):
        assert b1.is_similar_to(b2)
    for c1, c2 in zip(nproc1.crystals(), nproc4.crystals()):
        assert c1.is_similar_to(c2)
    for d1, d2 in zip(nproc1.detectors(), nproc4.detectors()):
        assert d1.is_similar_to(
            d2,
            fast_axis_tolerance=5e-5,
            slow_axis_tolerance=5e-5,
            origin_tolerance=5e-5,
        )


def test_restrained_refinement_with_fixed_parameterisations(dials_regression, tmp_path):
    # Avoid a regression to https://github.com/dials/dials/issues/1142 by
    # testing that refinement succeeds when some parameterisations are fixed
    # by parameter auto reduction code, but restraints are requested for
    # those parameterisations.

    # The phil scope
    from dials.algorithms.refinement.refiner import phil_scope

    user_phil = phil.parse(
        """
refinement {
  parameterisation {
    auto_reduction {
      min_nref_per_parameter = 90
      action = fail *fix remove
    }
    crystal {
      unit_cell {
        restraints {
          tie_to_target {
            values = 95 95 132 90 90 120
            sigmas = 1 1 1 0 0 0
            id = 0 1 2 3 4 5 6 7 8 9
          }
        }
      }
    }
  }
}
"""
    )

    working_phil = phil_scope.fetch(source=user_phil)
    working_params = working_phil.extract()

    # use the multi stills test data
    data_dir = Path(dials_regression) / "refinement_test_data" / "multi_stills"
    experiments_path = data_dir / "combined_experiments.json"
    pickle_path = data_dir / "combined_reflections.pickle"

    experiments = ExperimentListFactory.from_json_file(
        experiments_path, check_format=False
    )
    reflections = flex.reflection_table.from_file(pickle_path)

    refiner = RefinerFactory.from_parameters_data_experiments(
        working_params, reflections, experiments
    )

    history = refiner.run()
    rmsd_limits = (0.2044, 0.2220, 0.0063)
    for a, b in zip(history["rmsd"][-1], rmsd_limits):
        assert a < b



from __future__ import annotations

import collections
import copy
import json
import logging
import math
import random
import sys

import iotbx.phil
from cctbx import sgtbx, uctbx
from cctbx.sgtbx.bravais_types import bravais_lattice
from cctbx.sgtbx.lattice_symmetry import metric_subgroups
from dxtbx.model import ExperimentList
from libtbx import Auto

import dials.util
from dials.algorithms.merging.merge import prepare_merged_reflection_table
from dials.algorithms.symmetry import resolution_filter_from_reflections_experiments
from dials.algorithms.symmetry.absences.laue_groups_info import (
    laue_groups as laue_groups_for_absence_analysis,
)
from dials.algorithms.symmetry.absences.run_absences_checks import (
    run_systematic_absences_checks,
)
from dials.algorithms.symmetry.absences.screw_axes import ScrewAxisObserver
from dials.algorithms.symmetry.laue_group import LaueGroupAnalysis
from dials.array_family import flex
from dials.command_line.reindex import reindex_experiments
from dials.util import log, tabulate
from dials.util.exclude_images import (
    exclude_image_ranges_from_scans,
    get_selection_for_valid_image_ranges,
)
from dials.util.filter_reflections import filtered_arrays_from_experiments_reflections
from dials.util.multi_dataset_handling import (
    assign_unique_identifiers,
    parse_multiple_datasets,
    update_imageset_ids,
)
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.symmetry")

phil_scope = iotbx.phil.parse(
    """\
include scope dials.util.exclude_images.phil_scope
d_min = Auto
  .type = float(value_min=0)

min_i_mean_over_sigma_mean = 4
  .type = float(value_min=0)

min_cc_half = 0.6
  .type = float(value_min=0, value_max=1)

normalisation = kernel quasi ml_iso *ml_aniso
  .type = choice

lattice_group = None
  .type = space_group

seed = 230
  .type = int(value_min=0)

lattice_symmetry_max_delta = 2.0
  .type = float(value_min=0)

relative_length_tolerance = 0.05
  .type = float(value_min=0)

absolute_angle_tolerance = 2
  .type = float(value_min=0)

partiality_threshold = 0.4
  .type = float
  .help = "Use only reflections with a partiality above this threshold."

laue_group = auto
  .type = space_group
  .help = "Optionally specify the Laue group. If set to auto, then test all possible "
          "Laue groups. If set to None, then take the Laue group from the input file."

change_of_basis_op = None
  .type = str

best_monoclinic_beta = True
  .type = bool
  .help = "If True, then for monoclinic centered cells, I2 will be preferred over C2 if"
          "it gives a less oblique cell (i.e. smaller beta angle)."

systematic_absences {

  check = True
    .type = bool
    .help = "Check systematic absences for the current laue group."

  method = *direct fourier
    .type = choice
    .help = "Use fourier analysis or direct analysis of I/sigma to determine"
            "likelihood of systematic absences"

  significance_level = *0.95 0.975 0.99
    .type = choice
    .help = "Signficance to use when testing whether axial reflections are "
            "different to zero (absences and reflections in reflecting condition)."

}

output {
  log = dials.symmetry.log
    .type = str
  experiments = "symmetrized.expt"
    .type = path
  reflections = "symmetrized.refl"
    .type = path
  json = dials.symmetry.json
    .type = path
  html = "dials.symmetry.html"
    .type = path
    .help = "Filename for html report."
}
""",
    process_includes=True,
)


def median_unit_cell(experiments):
    uc_params = [flex.double() for i in range(6)]
    for c in experiments.crystals():
        for i, p in enumerate(c.get_unit_cell().parameters()):
            uc_params[i].append(p)
    return uctbx.unit_cell(parameters=[flex.median(p) for p in uc_params])


def unit_cells_are_similar_to(
    experiments, unit_cell, relative_length_tolerance, absolute_angle_tolerance
):
    return all(
        expt.crystal.get_unit_cell().is_similar_to(
            unit_cell,
            relative_length_tolerance=relative_length_tolerance,
            absolute_angle_tolerance=absolute_angle_tolerance,
        )
        for expt in experiments
    )


def change_of_basis_ops_to_minimum_cell(
    experiments, max_delta, relative_length_tolerance, absolute_angle_tolerance
):
    """
    Compute change of basis ops to map experiments to the minimum cell

    Map to the minimum cell via the best cell, which appears to guarantee that the
    resulting minimum cells are consistent.

    Args:
        experiments (ExperimentList): a list of experiments.
        reflections (list): a list of reflection tables

    Returns: The experiments and reflections mapped to the minimum cell
    """

    median_cell = median_unit_cell(experiments)
    unit_cells_are_similar = unit_cells_are_similar_to(
        experiments, median_cell, relative_length_tolerance, absolute_angle_tolerance
    )
    centring_symbols = [
        bravais_lattice(group=expt.crystal.get_space_group()).centring_symbol
        for expt in experiments
    ]
    if unit_cells_are_similar and len(set(centring_symbols)) == 1:
        groups = metric_subgroups(
            experiments[0]
            .crystal.get_crystal_symmetry()
            .customized_copy(unit_cell=median_cell),
            max_delta,
            enforce_max_delta_for_generated_two_folds=True,
        )
        group = groups.result_groups[0]
        cb_op_best_to_min = group["best_subsym"].change_of_basis_op_to_minimum_cell()
        cb_ops = [cb_op_best_to_min * group["cb_op_inp_best"]] * len(experiments)
    else:
        groups = [
            metric_subgroups(
                expt.crystal.get_crystal_symmetry(),
                max_delta,
                best_monoclinic_beta=False,
                enforce_max_delta_for_generated_two_folds=True,
            )
            for expt in experiments
        ]
        counter = collections.Counter(
            g.result_groups[0]["best_subsym"].space_group() for g in groups
        )
        target_group = counter.most_common()[0][0]
        cb_ops = []
        for expt in experiments:
            groups = metric_subgroups(
                expt.crystal.get_crystal_symmetry(),
                max_delta,
                best_monoclinic_beta=False,
                enforce_max_delta_for_generated_two_folds=True,
            )
            group = None
            for g in groups.result_groups:
                if g["best_subsym"].space_group() == target_group:
                    group = g
            if group:
                cb_ops.append(group["cb_op_inp_best"])
            else:
                cb_ops.append(None)
                logger.info(
                    f"Couldn't match unit cell to target symmetry:\n"
                    f"{expt.crystal.get_crystal_symmetry()}\n"
                    f"Target symmetry: {target_group.info()}"
                )
        ref_expts = ExperimentList(
            [expt for expt, cb_op in zip(experiments, cb_ops) if cb_op]
        ).change_basis(list(filter(None, cb_ops)))
        cb_op_ref_min = (
            ref_expts[0]
            .crystal.get_crystal_symmetry()
            .customized_copy(unit_cell=median_unit_cell(ref_expts))
            .change_of_basis_op_to_minimum_cell()
        )
        cb_ops = [cb_op_ref_min * cb_op if cb_op else None for cb_op in cb_ops]
    return cb_ops


def apply_change_of_basis_ops(experiments, reflections, change_of_basis_ops):
    """
    Apply the given change of basis ops to the input experiments and reflections

    Args:
        experiments (ExperimentList): a list of experiments.
        reflections (list): a list of reflection tables
        change_of_basis_ops (list): a list of cctbx.sgtbx.change_of_basis_op

    Returns: The experiments and reflections after application of the change of basis ops
    """

    for expt, refl, cb_op_inp_min in zip(experiments, reflections, change_of_basis_ops):
        refl["miller_index"] = cb_op_inp_min.apply(refl["miller_index"])
        expt.crystal = expt.crystal.change_basis(cb_op_inp_min)
        expt.crystal.set_space_group(sgtbx.space_group())
    return experiments, reflections


def eliminate_sys_absent(experiments, reflections):
    for i, expt in enumerate(experiments):
        if expt.crystal.get_space_group().n_ltr() > 1:
            effective_group = (
                expt.crystal.get_space_group().build_derived_reflection_intensity_group(
                    anomalous_flag=True
                )
            )
            sys_absent_flags = effective_group.is_sys_absent(
                reflections[i]["miller_index"]
            )
            if sys_absent_flags.count(True):
                reflections[i] = reflections[i].select(~sys_absent_flags)
                logger.info(
                    "Eliminating %i systematic absences for experiment %s",
                    sys_absent_flags.count(True),
                    expt.identifier,
                )
    return reflections


def get_subset_for_symmetry(experiments, reflection_tables, exclude_images=None):
    """Select an image range for symmetry analysis, or just select
    the first 360 degrees of data."""
    refls_for_sym = []
    if exclude_images:
        experiments = exclude_image_ranges_from_scans(
            reflection_tables, experiments, exclude_images
        )
        for refl, exp in zip(reflection_tables, experiments):
            sel = get_selection_for_valid_image_ranges(refl, exp)
            refls_for_sym.append(refl.select(sel))
    else:
        for expt, refl in zip(experiments, reflection_tables):
            sel = get_selection_for_valid_image_ranges(refl, expt)
            if not sel.count(False):
                # Use first 360 degrees if <360 deg i.e. first measured data,
                # but only if no reflections have been explicitly excluded
                # already
                scan_end = int(math.ceil(360 / abs(expt.scan.get_oscillation()[1])))
                if scan_end < len(expt.scan):
                    sel = refl["xyzobs.px.value"].parts()[2] <= scan_end
            refls_for_sym.append(refl.select(sel))
    return refls_for_sym


def symmetry(experiments, reflection_tables, params=None):
    """
    Run symmetry analysis

    Args:
        experiments: An experiment list.
        reflection_tables: A list of reflection tables.
        params: The dials.symmetry phil scope.
    """
    result = None
    if params is None:
        params = phil_scope.extract()
    refls_for_sym = []

    if params.laue_group is Auto:
        logger.info("=" * 80)
        logger.info("")
        logger.info("Performing Laue group analysis")
        logger.info("")

        # Transform models into miller arrays
        n_datasets = len(experiments)

        # Map experiments and reflections to minimum cell
        # Eliminate reflections that are systematically absent due to centring
        # of the lattice, otherwise they would lead to non-integer miller indices
        # when reindexing to a primitive setting
        cb_ops = change_of_basis_ops_to_minimum_cell(
            experiments,
            params.lattice_symmetry_max_delta,
            params.relative_length_tolerance,
            params.absolute_angle_tolerance,
        )
        reflection_tables = eliminate_sys_absent(experiments, reflection_tables)
        experiments, reflection_tables = apply_change_of_basis_ops(
            experiments, reflection_tables, cb_ops
        )

        refls_for_sym = get_subset_for_symmetry(
            experiments, reflection_tables, params.exclude_images
        )

        datasets = filtered_arrays_from_experiments_reflections(
            experiments,
            refls_for_sym,
            outlier_rejection_after_filter=True,
            partiality_threshold=params.partiality_threshold,
        )
        if len(datasets) != n_datasets:
            raise ValueError(
                """Some datasets have no reflection after prefiltering, please check
    input data and filtering settings e.g partiality_threshold"""
            )

        datasets = [
            ma.as_anomalous_array().merge_equivalents().array() for ma in datasets
        ]
        result = LaueGroupAnalysis(
            datasets,
            normalisation=params.normalisation,
            d_min=params.d_min,
            min_i_mean_over_sigma_mean=params.min_i_mean_over_sigma_mean,
            lattice_symmetry_max_delta=params.lattice_symmetry_max_delta,
            relative_length_tolerance=params.relative_length_tolerance,
            absolute_angle_tolerance=params.absolute_angle_tolerance,
            best_monoclinic_beta=params.best_monoclinic_beta,
        )
        logger.info("")
        logger.info(result)

        if params.output.json is not None:
            d = result.as_dict()
            d["cb_op_inp_min"] = [str(cb_op) for cb_op in cb_ops]
            # Copy the "input_symmetry" to "min_cell_symmetry" as it isn't technically
            # the input symmetry to dials.symmetry
            d["min_cell_symmetry"] = d["input_symmetry"]
            del d["input_symmetry"]
            json_str = json.dumps(d, indent=2)
            with open(params.output.json, "w") as f:
                f.write(json_str)

        # Change of basis operator from input unit cell to best unit cell
        cb_op_inp_best = result.best_solution.subgroup["cb_op_inp_best"]
        # Get the best space group.
        best_subsym = result.best_solution.subgroup["best_subsym"]
        best_space_group = best_subsym.space_group().build_derived_acentric_group()
        logger.info(
            tabulate(
                [[str(best_subsym.space_group_info()), str(best_space_group.info())]],
                ["Patterson group", "Corresponding MX group"],
            )
        )
        # Reindex the input data
        _, refls_for_sym = _reindex_experiments_reflections(
            experiments, refls_for_sym, best_space_group, cb_op_inp_best
        )
        experiments, reflection_tables = _reindex_experiments_reflections(
            experiments, reflection_tables, best_space_group, cb_op_inp_best
        )

    elif params.laue_group is not None:
        if params.change_of_basis_op is not None:
            cb_op = sgtbx.change_of_basis_op(params.change_of_basis_op)
        else:
            cb_op = sgtbx.change_of_basis_op()
        # Reindex the input data
        experiments, reflection_tables = _reindex_experiments_reflections(
            experiments, reflection_tables, params.laue_group.group(), cb_op
        )

    if params.systematic_absences.check:
        logger.info("=" * 80)
        logger.info("")
        logger.info("Analysing systematic absences")
        logger.info("")

        # Get the laue class from the current space group.
        space_group = experiments[0].crystal.get_space_group()
        laue_group = str(space_group.build_derived_patterson_group().info())
        logger.info("Laue group: %s", laue_group)
        if laue_group in ("I m -3", "I m m m"):
            if laue_group == "I m -3":
                logger.info(
                    """Space groups I 2 3 & I 21 3 cannot be distinguished with systematic absence
analysis, due to lattice centering.
Using space group I 2 3, space group I 21 3 is equally likely.\n"""
                )
            if laue_group == "I m m m":
                logger.info(
                    """Space groups I 2 2 2 & I 21 21 21 cannot be distinguished with systematic absence
analysis, due to lattice centering.
Using space group I 2 2 2, space group I 21 21 21 is equally likely.\n"""
                )
        elif laue_group not in laue_groups_for_absence_analysis:
            logger.info("No absences to check for this laue group\n")
        else:
            if not refls_for_sym:
                refls_for_sym = get_subset_for_symmetry(
                    experiments, reflection_tables, params.exclude_images
                )

            if (params.d_min is Auto) and (result is not None):
                d_min = result.intensities.resolution_range()[1]
            elif params.d_min is Auto:
                d_min = resolution_filter_from_reflections_experiments(
                    refls_for_sym,
                    experiments,
                    params.min_i_mean_over_sigma_mean,
                    params.min_cc_half,
                )
            else:
                d_min = params.d_min

            # combine before sys abs test - only triggers if laue_group=None and
            # multiple input files.
            if len(reflection_tables) > 1:
                joint_reflections = flex.reflection_table()
                for table in refls_for_sym:
                    joint_reflections.extend(table)
            else:
                joint_reflections = refls_for_sym[0]

            merged_reflections = prepare_merged_reflection_table(
                experiments,
                joint_reflections,
                d_min,
                partiality_threshold=params.partiality_threshold,
            )
            run_systematic_absences_checks(
                experiments,
                merged_reflections,
                float(params.systematic_absences.significance_level),
                method=params.systematic_absences.method,
            )

    logger.info(
        "Saving reindexed experiments to %s in space group %s",
        params.output.experiments,
        str(experiments[0].crystal.get_space_group().info()),
    )
    experiments.as_file(params.output.experiments)
    if params.output.reflections is not None:
        if len(reflection_tables) > 1:
            joint_reflections = flex.reflection_table()
            for table in reflection_tables:
                joint_reflections.extend(table)
        else:
            joint_reflections = reflection_tables[0]
        logger.info(
            "Saving %s reindexed reflections to %s",
            len(joint_reflections),
            params.output.reflections,
        )
        joint_reflections.as_file(params.output.reflections)

    if params.output.html and params.systematic_absences.check:
        ScrewAxisObserver().generate_html_report(params.output.html)


def _reindex_experiments_reflections(experiments, reflections, space_group, cb_op):
    """Reindex the input data."""
    reindexed_experiments = reindex_experiments(
        experiments, cb_op, space_group=space_group
    )
    reindexed_reflections = flex.reflection_table()
    reflections = update_imageset_ids(experiments, reflections)
    for i in range(len(reindexed_experiments)):
        reindexed_refl = copy.deepcopy(reflections[i])
        reindexed_refl["miller_index"] = cb_op.apply(reindexed_refl["miller_index"])
        reindexed_reflections.extend(reindexed_refl)
    return reindexed_experiments, [reindexed_reflections]


help_message = """
This program implements the methods of
`POINTLESS <http://www.ccp4.ac.uk/html/pointless.html>`_ (
`Evans, P. (2006). Acta Cryst. D62, 72-82. <https://doi.org/10.1107/S0907444905036693>`_ and
`Evans, P. R. (2011). Acta Cryst. D67, 282-292. <https://doi.org/10.1107/S090744491003982X>`_)
for scoring and determination of Laue group symmetry.

The program takes as input a set of one or more integrated experiments and
reflections.

Examples::

  dials.symmetry models.expt observations.refl
"""


@dials.util.show_mail_handle_errors()
def run(args=None):
    """Run symmetry analysis from the command-line."""
    usage = "dials.symmetry [options] models.expt observations.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options, args = parser.parse_args(
        args=args, show_diff_phil=False, return_unhandled=True
    )

    # Configure the logging
    log.config(verbosity=options.verbose, logfile=params.output.log)

    logger.info(dials_version())

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    if params.seed is not None:
        flex.set_random_seed(params.seed)
        random.seed(params.seed)

    if not params.input.experiments or not params.input.reflections:
        parser.print_help()
        sys.exit()

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    reflections = parse_multiple_datasets(reflections)

    if len(experiments) != len(reflections):
        sys.exit(
            "Mismatched number of experiments and reflection tables found: %s & %s."
            % (len(experiments), len(reflections))
        )
    try:
        experiments, reflections = assign_unique_identifiers(experiments, reflections)
        symmetry(experiments, reflections, params=params)
    except ValueError as e:
        sys.exit(e)


if __name__ == "__main__":
    run()



from __future__ import annotations

from unittest.mock import MagicMock, Mock

import pytest

from dxtbx import flumpy
from dxtbx.model import Beam, Crystal, Detector, Experiment, Goniometer, Scan
from dxtbx.model.experiment_list import ExperimentList
from libtbx import phil
from scitbx import sparse

from dials.algorithms.scaling.basis_functions import RefinerCalculator
from dials.algorithms.scaling.parameter_handler import ScalingParameterManagerGenerator
from dials.algorithms.scaling.scaler import (
    MultiScaler,
    NullScaler,
    SingleScaler,
    TargetScaler,
    calc_sf_variances,
)
from dials.algorithms.scaling.scaler_factory import create_scaler
from dials.algorithms.scaling.scaling_library import create_scaling_model
from dials.algorithms.scaling.scaling_utilities import calculate_prescaling_correction
from dials.algorithms.scaling.target_function import ScalingTarget
from dials.array_family import flex
from dials.util.options import ArgumentParser


def side_effect_update_var(variances, intensities):
    """Side effect to mock configure reflection table
    call during initialisation."""
    return flex.double(range(1, len(variances) + 1))


@pytest.fixture
def mock_errormodel():
    """A mock error model."""
    em = MagicMock()
    em.refined_parameters = [1.0, 0.1]
    em.update_variances.side_effect = side_effect_update_var
    return em


@pytest.fixture
def mock_errormodel2():
    """A mock error model."""
    em = MagicMock()
    em.refined_parameters = [1.0, 0.1]
    em.update_variances.side_effect = side_effect_update_var
    # return_value = flex.double(range(2, 9))
    return em


def generated_exp(n=1):
    """Generate an experiment list with two experiments."""
    experiments = ExperimentList()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " C 2y",
    }
    crystal = Crystal.from_dict(exp_dict)
    scan = Scan(image_range=[0, 90], oscillation=[0.0, 1.0])
    beam = Beam(s0=(0.0, 0.0, 1.01))
    goniometer = Goniometer((1.0, 0.0, 0.0))
    detector = Detector()
    experiments.append(
        Experiment(
            beam=beam,
            scan=scan,
            goniometer=goniometer,
            detector=detector,
            crystal=crystal,
        )
    )
    experiments[0].identifier = "0"
    if n > 1:
        for i in range(n - 1):
            experiments.append(
                Experiment(
                    beam=beam,
                    scan=scan,
                    goniometer=goniometer,
                    detector=detector,
                    crystal=crystal,
                )
            )
            experiments[i + 1].identifier = str(i + 1)
    return experiments


def generated_param():
    """Generate a param phil scope."""
    phil_scope = phil.parse(
        """
      include scope dials.algorithms.scaling.scaling_options.phil_scope
      include scope dials.algorithms.scaling.model.model.model_phil_scope
      include scope dials.algorithms.scaling.scaling_refiner.scaling_refinery_phil_scope
  """,
        process_includes=True,
    )
    parser = ArgumentParser(phil=phil_scope, check_format=False)
    parameters, _ = parser.parse_args(args=[], quick_parse=True, show_diff_phil=False)
    parameters.model = "KB"
    return parameters


def generated_refl(id_=0):
    """Create a reflection table suitable for splitting into blocks."""
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double([1.0, 2.0, 3.0, 4.0, 500.0, 6.0, 2.0, 2.0])
    reflections["variance"] = flex.double(8, 1.0)
    reflections["miller_index"] = flex.miller_index(
        [
            (1, 0, 0),
            (2, 0, 0),
            (0, 0, 1),
            (2, 2, 2),
            (1, 0, 0),
            (2, 0, 0),
            (1, 0, 0),
            (1, 0, 0),
        ]
    )
    reflections["d"] = flex.double([0.8, 2.1, 2.0, 1.4, 1.6, 2.5, 2.5, 2.5])
    reflections["partiality"] = flex.double(8, 1.0)
    reflections["Esq"] = flex.double(8, 1.0)
    reflections["inverse_scale_factor"] = flex.double(8, 1.0)
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [
            (0.0, 0.0, 0.0),
            (0.0, 0.0, 5.0),
            (0.0, 0.0, 8.0),
            (0.0, 0.0, 10.0),
            (0.0, 0.0, 12.0),
            (0.0, 0.0, 15.0),
            (0.0, 0.0, 15.0),
            (0.0, 0.0, 15.0),
        ]
    )
    reflections["s1"] = flex.vec3_double(
        [
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
        ]
    )
    reflections.set_flags(flex.bool(8, True), reflections.flags.integrated)
    reflections.set_flags(
        flex.bool([False] * 5 + [True] + [False] * 2),
        reflections.flags.excluded_for_scaling,
    )
    reflections["id"] = flex.int(8, id_)
    reflections.experiment_identifiers()[id_] = str(id_)
    return reflections


def generated_refl_for_comb():
    """Create a reflection table suitable for splitting into blocks."""
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double([1.0, 2.0, 3.0, 4.0, 500.0, 6.0, 2.0, 2.0])
    reflections["variance"] = flex.double(8, 1.0)
    reflections["intensity.prf.value"] = flex.double(
        [1.0, 3.0, 3.0, 4.0, 50.0, 6.0, 3.0, 2.0]
    )
    reflections["intensity.prf.variance"] = flex.double(
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0]
    )
    reflections["intensity.sum.value"] = flex.double(
        [1.0, 4.0, 3.0, 4.0, 500.0, 6.0, 6.0, 2.0]
    )
    reflections["intensity.sum.variance"] = flex.double(8, 1.0)
    reflections["miller_index"] = flex.miller_index(
        [
            (1, 0, 0),
            (2, 0, 0),
            (0, 0, 1),
            (2, 2, 2),
            (1, 0, 0),
            (2, 0, 0),
            (1, 0, 0),
            (1, 0, 0),
        ]
    )
    reflections["d"] = flex.double([0.8, 2.1, 2.0, 1.4, 1.6, 2.5, 2.5, 2.5])
    reflections["partiality"] = flex.double(8, 1.0)
    reflections["Esq"] = flex.double(8, 1.0)
    reflections["inverse_scale_factor"] = flex.double(8, 1.0)
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [
            (0.0, 0.0, 0.0),
            (0.0, 0.0, 5.0),
            (0.0, 0.0, 8.0),
            (0.0, 0.0, 10.0),
            (0.0, 0.0, 12.0),
            (0.0, 0.0, 15.0),
            (0.0, 0.0, 15.0),
            (0.0, 0.0, 15.0),
        ]
    )
    reflections["s1"] = flex.vec3_double(
        [
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
        ]
    )
    reflections.set_flags(flex.bool(8, True), reflections.flags.integrated)
    reflections.set_flags(
        flex.bool([False] * 5 + [True] + [False] * 2), reflections.flags.bad_for_scaling
    )
    reflections["id"] = flex.int(8, 0)
    reflections.experiment_identifiers()[0] = "0"
    reflections = calculate_prescaling_correction(reflections)
    return reflections


def mock_scaling_component():
    """Mock scaling component to allow creation of a scaling model."""
    component = MagicMock()
    component.n_params = 2
    component.inverse_scales = [flex.double([0.9, 1.1])]
    component.derivatives = [sparse.matrix(2, 2)]
    component.derivatives[0][0, 0] = 0.5
    component.derivatives[0][1, 0] = 0.4
    return component


@pytest.fixture
def mock_apm():
    """mock parameter manager for testing var_cov_matrix setting."""
    apm = MagicMock()
    apm.var_cov_matrix = flex.double([2.0])
    apm.var_cov_matrix.reshape(flex.grid(1, 1))
    apm.n_active_params = 1
    apm.n_obs = [2]
    apm.derivatives = [sparse.matrix(1, 1)]
    apm.components_list = ["scale"]
    apm.components = {
        "scale": {"object": mock_scaling_component(), "n_params": 1, "start_idx": 0}
    }
    apm.constant_g_values = [flex.double(2, 1.0)]
    return apm


def test_SingleScaler_initialisation():
    """Test that all attributes are correctly set upon initialisation"""
    p, e, r = (generated_param(), generated_exp(), generated_refl())
    exp = create_scaling_model(p, e, r)
    p.reflection_selection.method = "use_all"
    # test initialised correctly
    scaler = SingleScaler(p, exp[0], r)
    assert (
        list(scaler.suitable_refl_for_scaling_sel) == [True] * 5 + [False] + [True] * 2
    )
    # all 7 of the suitable should be within the scaling_subset
    assert list(scaler.scaling_subset_sel) == [True] * 7
    # one of these is not in the scaling selection due to being an outlier.
    assert list(scaler.scaling_selection) == [True] * 4 + [False] + [True] * 2
    assert list(scaler.outliers) == [False] * 4 + [True] + [False] * 2
    assert scaler.n_suitable_refl == 7

    # check for correct setup of global_Ih_table
    # block selection is order to extract out from suitable_reflections
    assert scaler.global_Ih_table.size == 7
    assert list(scaler.global_Ih_table.blocked_data_list[0].intensities) == [
        3.0,
        1.0,
        500.0,
        2.0,
        2.0,
        2.0,
        4.0,
    ]
    block_selection = scaler.global_Ih_table.blocked_data_list[0].block_selections[0]
    assert list(block_selection) == [2, 0, 4, 5, 6, 1, 3]

    # check for correct setup of Ih_table
    assert scaler.Ih_table.size == 6
    assert list(scaler.Ih_table.blocked_data_list[0].intensities) == [
        3.0,
        1.0,
        2.0,
        2.0,
        2.0,
        4.0,
    ]
    block_selection = scaler.Ih_table.blocked_data_list[0].block_selections[0]
    assert list(block_selection) == [2, 0, 5, 6, 1, 3]

    # check for correct data/d_values in components
    d_suitable = r["d"].select(scaler.suitable_refl_for_scaling_sel)
    decay = scaler.experiment.scaling_model.components["decay"]
    # first check 'data' contains all suitable reflections
    assert list(decay.data["d"]) == list(d_suitable)
    # Now check 'd_values' (which will be used for minim.) matches Ih_table data
    assert list(decay.d_values[0]) == list(
        d_suitable.select(flumpy.from_numpy(block_selection))
    )

    # test make ready for scaling method
    # set some new outliers and check for updated datastructures
    outlier_list = [False] * 3 + [True] * 2 + [False] * 2
    scaler.outliers = flex.bool(outlier_list)
    scaler.make_ready_for_scaling(outlier=True)
    assert scaler.Ih_table.size == 5
    assert list(scaler.Ih_table.blocked_data_list[0].intensities) == [
        3.0,
        1.0,
        2.0,
        2.0,
        2.0,
    ]
    block_selection = scaler.Ih_table.blocked_data_list[0].block_selections[0]
    assert list(block_selection) == [2, 0, 5, 6, 1]
    assert list(decay.d_values[0]) == list(
        d_suitable.select(flumpy.from_numpy(block_selection))
    )

    # test set outliers
    assert list(r.get_flags(r.flags.outlier_in_scaling)) == [False] * 8
    scaler._set_outliers()
    assert list(r.get_flags(r.flags.outlier_in_scaling)) == outlier_list + [False]


def test_multiscaler_initialisation():
    """Unit tests for the MultiScalerBase class."""
    p, e = (generated_param(), generated_exp(2))
    r1 = generated_refl(id_=0)
    r1["intensity.sum.value"] = r1["intensity"]
    r1["intensity.sum.variance"] = r1["variance"]
    r2 = generated_refl(id_=1)
    r2["intensity.sum.value"] = r2["intensity"]
    r2["intensity.sum.variance"] = r2["variance"]
    exp = create_scaling_model(p, e, [r1, r2])
    singlescaler1 = create_scaler(p, [exp[0]], [r1])
    singlescaler2 = create_scaler(p, [exp[1]], [r2])

    multiscaler = MultiScaler([singlescaler1, singlescaler2])

    # check initialisation
    assert len(multiscaler.active_scalers) == 2
    assert multiscaler.active_scalers[0] == singlescaler1
    assert multiscaler.active_scalers[1] == singlescaler2

    # check for correct setup of global Ih table
    assert multiscaler.global_Ih_table.size == 14
    assert (
        list(multiscaler.global_Ih_table.blocked_data_list[0].intensities)
        == [3.0, 1.0, 500.0, 2.0, 2.0, 2.0, 4.0] * 2
    )
    block_selections = multiscaler.global_Ih_table.blocked_data_list[0].block_selections
    assert list(block_selections[0]) == [2, 0, 4, 5, 6, 1, 3]
    assert list(block_selections[1]) == [2, 0, 4, 5, 6, 1, 3]

    # check for correct setup of Ih_table
    assert multiscaler.Ih_table.size == 12
    assert (
        list(multiscaler.Ih_table.blocked_data_list[0].intensities)
        == [3.0, 1.0, 2.0, 2.0, 2.0, 4.0] * 2
    )
    block_selections = multiscaler.Ih_table.blocked_data_list[0].block_selections
    assert list(block_selections[0]) == [2, 0, 5, 6, 1, 3]
    assert list(block_selections[1]) == [2, 0, 5, 6, 1, 3]

    # check for correct data/d_values in components
    for i, scaler in enumerate(multiscaler.active_scalers):
        d_suitable = scaler.reflection_table["d"].select(
            scaler.suitable_refl_for_scaling_sel
        )
        decay = scaler.experiment.scaling_model.components["decay"]
        # first check 'data' contains all suitable reflections
        assert list(decay.data["d"]) == list(d_suitable)
        # Now check 'd_values' (which will be used for minim.) matches Ih_table data
        assert list(decay.d_values[0]) == list(
            d_suitable.select(flumpy.from_numpy(block_selections[i]))
        )


def test_targetscaler_initialisation():
    """Unit tests for the MultiScalerBase class."""
    p, e = (generated_param(), generated_exp(2))
    r1 = generated_refl(id_=0)
    p.reflection_selection.method = "intensity_ranges"

    r1["intensity.sum.value"] = r1["intensity"]
    r1["intensity.sum.variance"] = r1["variance"]
    r2 = generated_refl(id_=1)
    r2["intensity.sum.value"] = r2["intensity"]
    r2["intensity.sum.variance"] = r2["variance"]
    exp = create_scaling_model(p, e, [r1, r2])
    singlescaler1 = SingleScaler(p, exp[0], r1, for_multi=True)
    singlescaler2 = SingleScaler(p, exp[1], r2, for_multi=True)

    # singlescaler2.experiments.scaling_model.set_scaling_model_as_scaled()

    targetscaler = TargetScaler(
        scaled_scalers=[singlescaler1], unscaled_scalers=[singlescaler2]
    )

    # check initialisation
    assert len(targetscaler.active_scalers) == 1
    assert len(targetscaler.single_scalers) == 1
    assert targetscaler.active_scalers[0] == singlescaler2
    assert targetscaler.single_scalers[0] == singlescaler1

    # check for correct setup of global Ih table
    assert targetscaler.global_Ih_table.size == 7  # only for active scalers
    assert list(targetscaler.global_Ih_table.blocked_data_list[0].intensities) == [
        3.0,
        1.0,
        500.0,
        2.0,
        2.0,
        2.0,
        4.0,
    ]
    block_selections = targetscaler.global_Ih_table.blocked_data_list[
        0
    ].block_selections
    assert list(block_selections[0]) == [2, 0, 4, 5, 6, 1, 3]

    # check for correct setup of Ih_table
    assert targetscaler.Ih_table.size == 6
    assert list(targetscaler.Ih_table.blocked_data_list[0].intensities) == [
        3.0,
        1.0,
        2.0,
        2.0,
        2.0,
        4.0,
    ]
    block_selections = targetscaler.Ih_table.blocked_data_list[0].block_selections
    assert list(block_selections[0]) == [2, 0, 5, 6, 1, 3]

    # check for correct setup of target Ih_Table
    assert targetscaler.target_Ih_table.size == 6
    assert list(targetscaler.target_Ih_table.blocked_data_list[0].intensities) == [
        3.0,
        1.0,
        2.0,
        2.0,
        2.0,
        4.0,
    ]
    block_selections = targetscaler.target_Ih_table.blocked_data_list[
        0
    ].block_selections
    assert list(block_selections[0]) == [
        2,
        0,
        4,
        5,
        1,
        3,
    ]  # different as taget_Ih_table
    # not created with indices lists.

    block_selections = targetscaler.Ih_table.blocked_data_list[0].block_selections
    # check for correct data/d_values in components
    for i, scaler in enumerate(targetscaler.active_scalers):
        d_suitable = scaler.reflection_table["d"].select(
            scaler.suitable_refl_for_scaling_sel
        )
        decay = scaler.experiment.scaling_model.components["decay"]
        # first check 'data' contains all suitable reflections
        assert list(decay.data["d"]) == list(d_suitable)
        # Now check 'd_values' (which will be used for minim.) matches Ih_table data
        assert list(decay.d_values[0]) == list(
            d_suitable.select(flumpy.from_numpy(block_selections[i]))
        )

    # but shouldn't have updated other
    assert (
        targetscaler.single_scalers[0]
        .experiment.scaling_model.components["decay"]
        .d_values
        == []
    )


def test_SingleScaler_expand_scales_to_all_reflections(mock_apm):
    p, e, r = (generated_param(), generated_exp(), generated_refl())
    exp = create_scaling_model(p, e, r)
    p.reflection_selection.method = "use_all"
    scaler = SingleScaler(p, exp[0], r)
    # test expand to all reflections method. First check scales are all 1, then
    # update a component to simulate a minimisation result, then check that
    # scales are set only in all suitable reflections (as it may not be possible
    # to calculate scales for unsuitable reflections!)
    # Must also update the scales in the global_Ih_table
    assert list(scaler.reflection_table["inverse_scale_factor"]) == [1.0] * 8
    scaler.experiment.scaling_model.components["scale"].parameters = flex.double([2.0])
    scaler.expand_scales_to_all_reflections(calc_cov=False)
    assert (
        list(scaler.reflection_table["inverse_scale_factor"])
        == [2.0] * 5 + [1.0] + [2.0] * 2
    )
    assert (
        list(scaler.global_Ih_table.blocked_data_list[0].inverse_scale_factors)
        == [2.0] * 7
    )

    assert list(scaler.reflection_table["inverse_scale_factor_variance"]) == [0.0] * 8
    # now try again
    apm = Mock()
    apm.n_active_params = 2
    var_list = [1.0, 0.1, 0.1, 0.5]
    apm.var_cov_matrix = flex.double(var_list)
    apm.var_cov_matrix.reshape(flex.grid(2, 2))
    scaler.update_var_cov(apm)
    assert scaler.var_cov_matrix[0, 0] == var_list[0]
    assert scaler.var_cov_matrix[0, 1] == var_list[1]
    assert scaler.var_cov_matrix[1, 0] == var_list[2]
    assert scaler.var_cov_matrix[1, 1] == var_list[3]
    assert scaler.var_cov_matrix.non_zeroes == 4
    scaler.expand_scales_to_all_reflections(calc_cov=True)
    assert list(
        scaler.reflection_table["inverse_scale_factor_variance"]
    ) == pytest.approx(
        [2.53320, 1.07106, 1.08125, 1.23219, 1.15442, 0.0, 1.0448, 1.0448], 1e-4
    )

    # Second case - when var_cov_matrix is only part of full matrix.
    p, e, r = (generated_param(), generated_exp(), generated_refl())
    exp = create_scaling_model(p, e, r)
    p.reflection_selection.method = "use_all"
    scaler = SingleScaler(p, exp[0], r)
    apm = mock_apm
    scaler.update_var_cov(apm)
    assert scaler.var_cov_matrix.non_zeroes == 1
    assert scaler.var_cov_matrix[0, 0] == 2.0
    assert scaler.var_cov_matrix.n_cols == 2
    assert scaler.var_cov_matrix.n_rows == 2
    assert scaler.var_cov_matrix.non_zeroes == 1


def generated_refl_2(exclude_refl=True):
    """Generate a reflection table."""
    # these miller_idx/d_values don't make physical sense, but I didn't want to
    # have to write the tests for lots of reflections.
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double([1.0, 10.0, 100.0, 1.0])
    reflections["variance"] = flex.double([1.0, 10.0, 100.0, 1.0])
    reflections["miller_index"] = flex.miller_index(
        [(1, 0, 0), (0, 0, 1), (2, 0, 0), (2, 2, 2)]
    )  # don't change
    reflections["d"] = flex.double([0.8, 2.0, 2.0, 0.0])  # don't change
    reflections["d"] = flex.double([0.8, 2.0, 2.1, 0.1])
    reflections["Esq"] = flex.double([1.0, 1.0, 1.0, 1.0])
    reflections["inverse_scale_factor"] = flex.double([1.0, 1.0, 1.0, 1.0])
    reflections["id"] = flex.int(4, 0)
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [(0.0, 0.0, 0.0), (0.0, 0.0, 5.0), (0.0, 0.0, 10.0), (0.0, 0.0, 10.0)]
    )
    reflections["s1"] = flex.vec3_double(
        [(0.0, 0.1, 1.0), (0.0, 0.1, 1.0), (0.0, 0.1, 1.0), (0.0, 0.1, 1.0)]
    )
    if exclude_refl:
        integrated_list = flex.bool([True, True, False, False])
        bad_list = flex.bool([False, False, True, True])
    else:
        integrated_list = flex.bool(4, True)
        bad_list = flex.bool(4, False)
    reflections.set_flags(integrated_list, reflections.flags.integrated)
    reflections.set_flags(bad_list, reflections.flags.bad_for_scaling)
    return reflections


def test_SingleScaler_update_for_minimisation():
    """Test the update_for_minimisation method of the singlescaler."""
    # test_params.scaling_options.nproc = 1
    p, e, r = (generated_param(), generated_exp(), generated_refl_2())
    exp = create_scaling_model(p, e, r)
    p.reflection_selection.method = "use_all"
    single_scaler = SingleScaler(p, exp[0], r)
    pmg = ScalingParameterManagerGenerator(
        single_scaler.active_scalers,
        ScalingTarget(),
        single_scaler.params.scaling_refinery.refinement_order,
    )
    single_scaler.components["scale"].parameters /= 2.0
    apm = pmg.parameter_managers()[0]

    Ih_table = single_scaler.Ih_table.blocked_data_list[0]
    Ih_table.calc_Ih()
    assert list(Ih_table.inverse_scale_factors) == [1.0, 1.0]
    assert list(Ih_table.Ih_values) == [10.0, 1.0]
    single_scaler.update_for_minimisation(apm, 0)
    # Should set new scale factors, and calculate Ih and weights.
    bf = RefinerCalculator.calculate_scales_and_derivatives(apm.apm_list[0], 0)
    assert list(Ih_table.inverse_scale_factors) == list(bf[0])
    assert list(Ih_table.Ih_values) != [1.0, 10.0]
    assert list(Ih_table.Ih_values) == pytest.approx(list(Ih_table.intensities / bf[0]))
    for i in range(Ih_table.derivatives.n_rows):
        for j in range(Ih_table.derivatives.n_cols):
            assert Ih_table.derivatives[i, j] == pytest.approx(bf[1][i, j])
    assert Ih_table.derivatives.non_zeroes == bf[1].non_zeroes


# @pytest.mark.xfail(reason='need to rework mcok error model')
def test_update_error_model(mock_errormodel, mock_errormodel2):
    """Test the update_error_model method"""
    p, e, r = (generated_param(), generated_exp(), generated_refl())
    exp = create_scaling_model(p, e, r)
    p.reflection_selection.method = "use_all"
    # test initialised correctly
    scaler = SingleScaler(p, exp[0], r)
    block = scaler.global_Ih_table.blocked_data_list[0]
    original_vars = block.variances
    # test update error model - should update weights in global Ih
    # as will be setting different things in Ih_table and reflection table, split
    # up the test to use two different error models.
    scaler._update_error_model(mock_errormodel)
    assert list(block.variances) == list(original_vars)
    newvars = flex.double(range(1, 8))
    assert list(block.block_selections[0]) == [2, 0, 4, 5, 6, 1, 3]
    assert list(block.weights) == list(1.0 / newvars)
    assert scaler.experiment.scaling_model.error_model is mock_errormodel

    # now test for updating of reflection table
    # do again with second errormodel
    scaler._update_error_model(mock_errormodel2)
    assert list(block.variances) == list(original_vars)
    newvars = flex.double(range(1, 9))
    assert list(block.block_selections[0]) == [2, 0, 4, 5, 6, 1, 3]
    # [2, 3, 4, 5, 6, 7, 8] < set these in ^ these positions (taking into account
    # the one non-suitable refl at index 5)
    assert list(block.weights) == list(1.0 / newvars)[:-1]
    assert scaler.experiment.scaling_model.error_model is mock_errormodel2


def test_SingleScaler_combine_intensities():
    """test combine intensities method"""
    p, e, r = (generated_param(), generated_exp(), generated_refl_for_comb())
    exp = create_scaling_model(p, e, r)
    p.reflection_selection.method = "use_all"
    scaler = SingleScaler(p, exp[0], r)
    scaler.combine_intensities()

    # The input makes the profile intensities best - so check these are set in the
    # reflection table and global_Ih_table
    assert list(scaler.reflection_table["intensity"]) == list(r["intensity.prf.value"])
    assert list(scaler.reflection_table["variance"]) == list(
        r["intensity.prf.variance"]
    )
    block = scaler.global_Ih_table.blocked_data_list[0]
    block_sel = flumpy.from_numpy(block.block_selections[0])
    suitable = scaler.suitable_refl_for_scaling_sel
    assert list(block.intensities) == list(
        scaler.reflection_table["intensity"].select(suitable).select(block_sel)
    )
    assert list(block.variances) == list(
        scaler.reflection_table["variance"].select(suitable).select(block_sel)
    )


def test_NullScaler():
    """Test for successful creation of NullScaler."""
    p, e, r = (generated_param(), generated_exp(), generated_refl())
    exp = create_scaling_model(p, e, r)
    _ = NullScaler(p, exp[0], r)
    # What exactly should be tested here?


def test_sf_variance_calculation():
    """Test the calculation of scale factor variances."""
    test_experiments = generated_exp()
    test_params = generated_param()
    assert len(test_experiments) == 1
    experiments = create_scaling_model(test_params, test_experiments, [None])
    components = experiments[0].scaling_model.components
    rt = flex.reflection_table()
    d1 = 1.0
    d2 = 2.0
    d3 = 3.0
    rt["d"] = flex.double([d1, d2, d3])
    rt["id"] = flex.int([0, 0, 0])
    experiments[0].scaling_model.configure_components(rt, experiments[0], test_params)
    components["scale"].update_reflection_data()
    _, d = components["scale"].calculate_scales_and_derivatives()
    assert list(d.col(0)) == [(0, 1.0), (1, 1.0), (2, 1.0)]
    components["decay"].update_reflection_data()
    s, d = components["decay"].calculate_scales_and_derivatives()
    assert list(d.col(0)) == [
        (0, 1.0 / (2.0 * d1 * d1)),
        (1, 1.0 / (2.0 * d2 * d2)),
        (2, 1.0 / (2.0 * d3 * d3)),
    ]
    var_cov = sparse.matrix(2, 2)
    a = 0.2
    b = 0.3
    c = 0.1
    var_cov[0, 0] = a
    var_cov[0, 1] = c
    var_cov[1, 0] = c
    var_cov[1, 1] = b
    variances = calc_sf_variances(components, var_cov)
    assert list(variances) == pytest.approx(
        [
            b / (4.0 * (d1**4.0)) + c / (d1**2.0) + a,
            b / (4.0 * (d2**4.0)) + c / (d2**2.0) + a,
            b / (4.0 * (d3**4.0)) + c / (d3**2.0) + a,
        ]
    )


def test_multiscaler_update_for_minimisation():
    """Test the multiscaler update_for_minimisation method."""

    p, e = (generated_param(), generated_exp(2))
    p.reflection_selection.method = "use_all"
    r1 = generated_refl(id_=0)
    r1["intensity.sum.value"] = r1["intensity"]
    r1["intensity.sum.variance"] = r1["variance"]
    r2 = generated_refl(id_=1)
    r2["intensity.sum.value"] = r2["intensity"]
    r2["intensity.sum.variance"] = r2["variance"]
    p.scaling_options.nproc = 2
    p.model = "physical"
    exp = create_scaling_model(p, e, [r1, r2])
    singlescaler1 = create_scaler(p, [exp[0]], [r1])
    singlescaler2 = create_scaler(p, [exp[1]], [r2])

    multiscaler = MultiScaler([singlescaler1, singlescaler2])
    pmg = ScalingParameterManagerGenerator(
        multiscaler.active_scalers,
        ScalingTarget,
        multiscaler.params.scaling_refinery.refinement_order,
    )
    multiscaler.single_scalers[0].components["scale"].parameters /= 2.0
    multiscaler.single_scalers[1].components["scale"].parameters *= 1.5
    apm = pmg.parameter_managers()[0]
    multiscaler.update_for_minimisation(apm, 0)
    multiscaler.update_for_minimisation(apm, 1)
    # bf[0], bf[1] should be list of scales and derivatives
    s1, d1 = RefinerCalculator.calculate_scales_and_derivatives(apm.apm_list[0], 0)
    s2, d2 = RefinerCalculator.calculate_scales_and_derivatives(apm.apm_list[1], 0)
    s3, d3 = RefinerCalculator.calculate_scales_and_derivatives(apm.apm_list[0], 1)
    s4, d4 = RefinerCalculator.calculate_scales_and_derivatives(apm.apm_list[1], 1)
    expected_scales_for_block_1 = s1
    expected_scales_for_block_1.extend(s2)
    expected_scales_for_block_2 = s3
    expected_scales_for_block_2.extend(s4)

    expected_derivatives_for_block_1 = sparse.matrix(
        expected_scales_for_block_1.size(), apm.n_active_params
    )
    expected_derivatives_for_block_2 = sparse.matrix(
        expected_scales_for_block_2.size(), apm.n_active_params
    )

    expected_derivatives_for_block_1.assign_block(d1, 0, 0)
    expected_derivatives_for_block_1.assign_block(
        d2, d1.n_rows, apm.apm_data[1]["start_idx"]
    )
    expected_derivatives_for_block_2.assign_block(d3, 0, 0)
    expected_derivatives_for_block_2.assign_block(
        d4, d3.n_rows, apm.apm_data[1]["start_idx"]
    )

    block_list = multiscaler.Ih_table.blocked_data_list

    assert list(block_list[0].inverse_scale_factors) == list(
        expected_scales_for_block_1
    )
    assert list(block_list[1].inverse_scale_factors) == list(
        expected_scales_for_block_2
    )
    assert block_list[1].derivatives == expected_derivatives_for_block_2
    assert block_list[0].derivatives == expected_derivatives_for_block_1



from __future__ import annotations

import os
import subprocess
from pathlib import Path

# DIALS version numbers are constructed from
#  1. a common prefix
__dials_version_format = "DIALS %s"
#  2. the most recent annotated git tag (or failing that: a default string)
__dials_version_default = "3.12"
#  3. a dash followed by the number of commits since that tag
#  4. a dash followed by a lowercase 'g' and the current commit id


def get_git_version(
    dials_path: Path | str, treat_merges_as_single_commit=False
) -> str | None:
    dials_path = str(dials_path)
    version = None
    with open(os.devnull, "w") as devnull:
        # Obtain name of the current branch. If this fails then the other commands will probably also fail
        branch = (
            subprocess.check_output(
                ["git", "branch", "--all", "--contains", "HEAD"],
                cwd=dials_path,
                stderr=devnull,
            )
            .rstrip()
            .decode("latin-1")
        )
        releasebranch = "dials-3" in branch

        # Always treat merges as single commit on release branches
        if releasebranch:
            treat_merges_as_single_commit = True

        # Get descriptive version string, eg. v1.1.0-1-g56f9cd7
        if treat_merges_as_single_commit:
            try:
                # Get a 'correct' depth, which should be the shortest path to the most recent tag
                version = (
                    subprocess.check_output(
                        ["git", "describe", "--long", "--first-parent"],
                        cwd=dials_path,
                        stderr=devnull,
                    )
                    .rstrip()
                    .decode("latin-1")
                )
            except Exception:
                pass  # This is not supported on older git versions < 1.8.4.
        if version is None:
            # Find the most recent tag
            version = (
                subprocess.check_output(
                    ["git", "describe", "--long"], cwd=dials_path, stderr=devnull
                )
                .rstrip()
                .decode("latin-1")
            )
            if treat_merges_as_single_commit:
                tag = version[: version.rindex("-", 0, version.rindex("-"))]
                commit = version[version.rindex("-") + 1 :]  # 'gxxxxxxx'
                # Now find the first-parent-path
                depth = subprocess.check_output(
                    ["git", "rev-list", f"{tag}..HEAD", "--first-parent"],
                    cwd=dials_path,
                    stderr=devnull,
                ).rstrip()
                if depth:
                    depth = depth.strip().count("\n") + 1
                else:
                    depth = 0
                version = "%s-%d-%s" % (tag, depth, commit)

        # Turn descriptive version string into proper version number
        if version[0] == "v":
            version = version[1:].replace(".0-", "-")
        version = version.replace("-", ".", 1)
        # If we are on a release branch, then append a '-release'-tag
        if releasebranch:
            version = version + "-release"
    return str(version)


# When run from a development installation the version information is extracted
# from the git repository. Otherwise it is read from the file '.gitversion' in the
# DIALS module directory.


def dials_version() -> str:
    """Try to obtain the current git revision number
    and store a copy in .gitversion"""
    version = None

    try:
        dials_path = Path(__file__).resolve().parent.parent
        repo_root = dials_path.parent.parent  # dials is in SRC
        version_file = dials_path / ".gitversion"

        # 1. Try to access information in .git directory
        #    Regenerate .gitversion if possible
        if (
            not os.environ.get("DIALS_SKIP_GIT_VERSIONING")
            and (repo_root / ".git").is_dir()
        ):
            try:
                version = get_git_version(dials_path)
                version_file.write_text(version)
            except Exception:
                if version == "":
                    version = None

        # 2. If .git directory missing or 'git describe' failed, read .gitversion
        if (version is None) and version_file.is_file():
            version = version_file.read_text().rstrip()
    except Exception:
        pass

    if version is None:
        version = __dials_version_format % __dials_version_default
    else:
        version = __dials_version_format % version

    return version



from __future__ import annotations

from dials.algorithms.integration import TimingInfo


def test_timing_info():
    """Create TimingInfo objects, test addition and str methods."""
    time1 = TimingInfo()
    time2 = TimingInfo()

    time1.read = 5
    time1.process = 4
    time1.user = 10
    time2.read = 15
    time2.process = 6
    time2.user = 1

    total = time1 + time2
    assert total.read == 20
    assert total.process == 10
    assert total.user == 11

    s = str(total)
    assert (
        s
        == """+--------------+---------------+
| Read time    | 20.00 seconds |
| Process time | 10.00 seconds |
| User time    | 11.00 seconds |
+--------------+---------------+"""
    )



from __future__ import annotations

from scitbx import matrix

from dials.algorithms.refinement.parameterisation.beam_parameters import BeamMixin
from dials.algorithms.refinement.parameterisation.scan_varying_model_parameters import (
    GaussianSmoother,
    ScanVaryingModelParameterisation,
    ScanVaryingParameterSet,
)


class ScanVaryingBeamParameterisation(ScanVaryingModelParameterisation, BeamMixin):
    """A scan-varying parameterisation for the beam with angles expressed in
    mrad and wavenumber in inverse Angstroms."""

    def __init__(
        self, beam, t_range, num_intervals, goniometer=None, experiment_ids=None
    ):

        if experiment_ids is None:
            experiment_ids = [0]

        # The state of a scan varying beam parameterisation is a vector
        # '[s0](t)', expressed as a function of image number 't'
        # in a sequential scan.
        #
        # The initial state is a snapshot of the beam unit direction
        # at the point of initialisation '[s0dir]', which is independent of
        # image number.
        #
        # Future states are composed by rotations around axes orthogonal to the
        # initial direction and scaling by the wavenumber.
        #
        # [s0](t) = nu(t) * [Mu2](t)[Mu1](t)[s0dir]

        # Set up the smoother
        smoother = GaussianSmoother(t_range, num_intervals)
        nv = smoother.num_values()

        # Set up the initial state
        s0 = matrix.col(beam.get_s0())
        istate = {
            "unit_s0": matrix.col(beam.get_unit_s0()),
            "polarization_normal": matrix.col(beam.get_polarization_normal()),
        }
        self._s0_at_t = s0

        # Factory function to provide to _build_p_list
        def parameter_type(value, axis, ptype, name):
            return ScanVaryingParameterSet(value, nv, axis, ptype, name)

        # Build the parameter list
        p_list = self._build_p_list(s0, goniometer, parameter_type=parameter_type)

        # Set up the base class
        ScanVaryingModelParameterisation.__init__(
            self, beam, istate, p_list, smoother, experiment_ids=experiment_ids
        )

        return

    def compose(self, t):
        """calculate state and derivatives for model at image number t"""

        # extract direction from the initial state
        ius0 = self._initial_state["unit_s0"]
        ipn = self._initial_state["polarization_normal"]

        # extract parameter sets from the internal list
        mu1_set, mu2_set, nu_set = self._param

        # extract data at time t using the smoother
        mu1, mu1_weights, mu1_sumweights = self._smoother.value_weight(t, mu1_set)
        mu2, mu2_weights, mu2_sumweights = self._smoother.value_weight(t, mu2_set)
        nu, nu_weights, nu_sumweights = self._smoother.value_weight(t, nu_set)

        # calculate derivatives of values at image t wrt underlying parameters.
        dmu1_dp = mu1_weights * (1.0 / mu1_sumweights)
        dmu2_dp = mu2_weights * (1.0 / mu2_sumweights)
        dnu_dp = nu_weights * (1.0 / nu_sumweights)

        # calculate new s0 and its derivatives wrt the values mu1, mu2 and nu
        (self._s0_at_t, _), ds0_dval = self._compose_core(
            ius0, ipn, mu1, mu2, nu, mu1_axis=mu1_set.axis, mu2_axis=mu2_set.axis
        )

        # calculate derivatives of state wrt underlying smoother parameters
        ds0_dp1 = [None] * dmu1_dp.size
        for (i, v) in dmu1_dp:
            ds0_dp1[i] = ds0_dval[0] * v
        ds0_dp2 = [None] * dmu2_dp.size
        for (i, v) in dmu2_dp:
            ds0_dp2[i] = ds0_dval[1] * v
        ds0_dp3 = [None] * dnu_dp.size
        for (i, v) in dnu_dp:
            ds0_dp3[i] = ds0_dval[2] * v

        # store derivatives as list-of-lists
        self._dstate_dp = [ds0_dp1, ds0_dp2, ds0_dp3]

        return

    def get_state(self):
        """Return beam vector [s0] at image number t"""

        # only a single beam is parameterised here, so no multi_state_elt
        # argument is allowed

        return self._s0_at_t



"""Setup experimental geometry for refinement test cases"""


from __future__ import annotations

import random

from dxtbx.model import BeamFactory, Crystal, DetectorFactory, GoniometerFactory
from libtbx.phil import command_line, parse
from scitbx import matrix


# Local functions
def random_vector_close_to(vector, sd=0.5):
    return matrix.col(vector).rotate_around_origin(
        matrix.col((random.random(), random.random(), random.random())).normalize(),
        random.gauss(0, sd),
        deg=True,
    )


class Extract:
    """Parse and extract geometry model from PHIL"""

    def __init__(
        self, master_phil, local_overrides="", cmdline_args=None, verbose=False
    ):

        self._verbose = verbose

        arg_interpreter = command_line.argument_interpreter(master_phil=master_phil)

        user_phil = parse(local_overrides)
        cmdline_phils = []
        if cmdline_args:
            for arg in cmdline_args:
                cmdline_phils.append(arg_interpreter.process(arg))

        working_phil = master_phil.fetch(sources=[user_phil] + cmdline_phils)

        self._params = working_phil.extract().geometry.parameters

        self.set_seed()

        self.build_goniometer()

        self.build_crystal()

        self.build_beam()

        self.build_detector()

        # write changes back to the PHIL object
        temp = working_phil.extract()
        temp.geometry.parameters = self._params
        self.phil = master_phil.format(python_object=temp)

    def set_seed(self):

        if self._params.random_seed is not None:
            random.seed(self._params.random_seed)
            # set the flex random seed too
            from dials.array_family import flex

            flex.set_random_seed(self._params.random_seed)
            if self._verbose:
                msg = "Random seed set to %d while building models"
                print(msg % self._params.random_seed)

    def build_goniometer(self):

        self.goniometer = GoniometerFactory.known_axis(self._params.goniometer.axis)

    def build_beam(self):

        if self._params.beam.wavelength.random:
            wavelength = random.uniform(*self._params.beam.wavelength.range)
        else:
            wavelength = self._params.beam.wavelength.value

        assert self._params.beam.direction.method in [
            "inclination",
            "close_to",
            "exactly",
        ]

        if self._params.beam.direction.method == "inclination":

            if self._params.beam.direction.inclination.random:
                inclination = random.gauss(
                    0.0, self._params.beam.direction.inclination.angle
                )
            else:
                inclination = self._params.beam.direction.inclination.angle

            beam_dir = matrix.col((0, 0, 1)).rotate_around_origin(
                matrix.col((0, 1, 0)), inclination, deg=True
            )

        elif self._params.beam.direction.method == "close_to":

            temp = self._params.beam.direction.close_to.direction
            beam_dir = random_vector_close_to(
                temp, sd=self._params.beam.direction.close_to.sd
            )

        elif self._params.beam.direction.method == "exactly":

            beam_dir = matrix.col(self._params.beam.direction.exactly)

        self.beam = BeamFactory.make_beam(unit_s0=beam_dir, wavelength=wavelength)

    def build_detector(self):

        assert self._params.detector.directions.method in ["close_to", "exactly"]

        if self._params.detector.directions.method == "close_to":

            temp = self._params.detector.directions.close_to.dir1
            dir1 = random_vector_close_to(
                temp, sd=self._params.detector.directions.close_to.sd
            )

            n = random_vector_close_to(
                self._params.detector.directions.close_to.norm,
                sd=self._params.detector.directions.close_to.sd,
            )

        elif self._params.detector.directions.method == "exactly":

            temp = self._params.detector.directions.exactly.dir1
            dir1 = matrix.col(temp)

            n = matrix.col(self._params.detector.directions.exactly.norm)

        dir2 = n.cross(dir1).normalize()

        assert self._params.detector.centre.method in ["close_to", "exactly"]

        if self._params.detector.centre.method == "close_to":

            centre = random_vector_close_to(
                self._params.detector.centre.close_to.value,
                sd=self._params.detector.centre.close_to.sd,
            )

        elif self._params.detector.centre.method == "exactly":

            temp = self._params.detector.centre.exactly.value
            centre = matrix.col(temp)

        origin = centre - (
            0.5 * self._params.detector.npx_fast * self._params.detector.pix_size * dir1
            + 0.5
            * self._params.detector.npx_slow
            * self._params.detector.pix_size
            * dir2
        )
        self.detector = DetectorFactory.make_detector(
            "PAD",
            dir1,
            dir2,
            origin,
            (self._params.detector.pix_size, self._params.detector.pix_size),
            (self._params.detector.npx_fast, self._params.detector.npx_slow),
            (0, 1.0e6),
        )

    @staticmethod
    def _build_cell_vec(vec):

        if vec.length.random:
            length = random.uniform(*vec.length.range)
        else:
            length = vec.length.value

        assert vec.direction.method in ["close_to", "exactly"]

        if vec.direction.method == "close_to":

            x = random_vector_close_to(
                vec.direction.close_to.direction, sd=vec.direction.close_to.sd
            )

        elif vec.direction.method == "exactly":

            x = matrix.col(vec.direction.exactly.direction)

        return length * x

    def build_crystal(self):

        vecs = [
            self._build_cell_vec(axis)
            for axis in (
                self._params.crystal.a,
                self._params.crystal.b,
                self._params.crystal.c,
            )
        ]

        sg = self._params.crystal.space_group_symbol

        self.crystal = Crystal(*vecs, space_group_symbol=sg)



#!/usr/bin/env python
#
# dials.ellipsoid.py
#
#  Copyright (C) 2018 Diamond Light Source
#
#  Author: James Parkhurst
#
#  This code is distributed under the BSD license, a copy of which is
#  included in the root directory of this package.

from __future__ import annotations

import copy
import logging
from math import pi

import numpy as np
from numpy.linalg import inv, norm

from dxtbx import flumpy
from libtbx.phil import parse
from scitbx import matrix

from dials.algorithms.profile_model.ellipsoid import chisq_pdf
from dials.algorithms.profile_model.ellipsoid.model import ProfileModelFactory
from dials.algorithms.profile_model.ellipsoid.parameterisation import ModelState
from dials.algorithms.profile_model.ellipsoid.refiner import Refiner as ProfileRefiner
from dials.algorithms.profile_model.ellipsoid.refiner import RefinerData
from dials.algorithms.profile_model.gaussian_rs import BBoxCalculator, MaskCalculator
from dials.algorithms.profile_model.gaussian_rs.calculator import (
    ComputeEsdBeamDivergence,
)
from dials.algorithms.spot_prediction import IndexGenerator
from dials.array_family import flex
from dials.constants import FULL_PARTIALITY

logger = logging.getLogger("dials")

ellipsoid_algorithm_phil_scope = parse(
    """
  integration {

    use_crude_shoebox_mask = False
      .type = bool

    shoebox {

      probability = %f
        .type = float

    }

    corrections {
      lp = True
        .type = bool
      dqe = True
        .type = bool
      partiality = True
        .type = bool
    }

  }

  debug {
    output {

      strong_spots = False
        .type = bool

      shoeboxes = False
        .type = bool

      profile_model = True
        .type = bool

      history = True
        .type = bool

      plots = False
        .type = bool

      print_shoeboxes = False
        .type = bool
    }
  }


"""
    % FULL_PARTIALITY
)

# Parameters
phil_scope = parse(
    """
include scope dials.algorithms.profile_model.ellipsoid.model.phil_scope
""",
    process_includes=True,
).adopt_scope(ellipsoid_algorithm_phil_scope)


def _compute_beam_vector(experiment, reflection_table):
    """Compute the obseved beam vector"""
    s1_obs = flex.vec3_double(len(reflection_table))
    for i in range(len(s1_obs)):
        x, y, _ = reflection_table["xyzobs.px.value"][i]
        s1_obs[i] = experiment.detector[0].get_pixel_lab_coord((x, y))
    return s1_obs


def _compute_bbox(experiment, reflection_table, sigma_d, s1="s1_obs"):
    """Compute the bounding box"""

    # Initialise the bounding box calculator
    compute_bbox = BBoxCalculator(
        experiment.crystal,
        experiment.beam,
        experiment.detector,
        experiment.goniometer,
        experiment.scan,
        sigma_d * 6,
        0,
    )

    # Compute the bounding box
    bbox = compute_bbox(
        reflection_table[s1],
        reflection_table["xyzcal.px"].parts()[2],
        reflection_table["panel"],
    )

    return bbox


def _compute_mask(
    experiment, reflection_table, sigma_d, s1="s1_obs", strong_shoeboxes=None
):
    """Compute the spot mask"""

    # Initialise the mask calculator
    mask_foreground = MaskCalculator(
        experiment.crystal,
        experiment.beam,
        experiment.detector,
        experiment.goniometer,
        experiment.scan,
        sigma_d * 3,
        0,
    )

    # Compute the reflection mask
    mask_foreground(
        reflection_table["shoebox"],
        reflection_table[s1],
        reflection_table["xyzcal.px"].parts()[2],
        reflection_table["panel"],
    )

    # if strong_shoeboxes given, apply the mask from this
    if strong_shoeboxes:
        # Apply strong spot mask
        assert len(reflection_table) == len(strong_shoeboxes)
        new_shoeboxes = reflection_table["shoebox"]
        for s in range(len(new_shoeboxes)):
            bbox_old = strong_shoeboxes[s].bbox
            mask_old = strong_shoeboxes[s].mask
            bbox_new = new_shoeboxes[s].bbox
            mask_new = new_shoeboxes[s].mask
            for j in range(mask_old.all()[1]):
                for i in range(mask_old.all()[2]):
                    ii = bbox_old[0] + i - bbox_new[0]
                    jj = bbox_old[2] + j - bbox_new[2]
                    if mask_old[0, j, i] == 5:
                        if (
                            ii >= 0
                            and jj >= 0
                            and jj < mask_new.all()[1]
                            and ii < mask_new.all()[2]
                        ):
                            assert mask_new[0, jj, ii] & (1 << 0)
                            mask_new[0, jj, ii] |= (1 << 2) | (1 << 3)


def initial_integrator(experiments, reflection_table):
    """Performs an initial integration of strong spots"""

    # some functions require an experimentlist, others just the experiment
    experiment = experiments[0]
    sel = reflection_table.get_flags(reflection_table.flags.strong)
    strong_refls = reflection_table.select(sel)
    strong_shoeboxes = copy.deepcopy(
        strong_refls["shoebox"]
    )  # Save the strong shoeboxes

    # Compute and initial spot size estimate and beam vector
    sigma_d = ComputeEsdBeamDivergence(experiment.detector, strong_refls).sigma()
    sigma_degrees = sigma_d * 180 / pi
    logger.info(
        f"Initial sigma d estimate for {len(strong_refls)} reflections\n"
        + f"Sigma D: {sigma_degrees:.5f} degrees\n",
    )
    strong_refls["s1_obs"] = _compute_beam_vector(experiment, strong_refls)
    strong_refls["bbox"] = _compute_bbox(experiment, strong_refls, sigma_d, "s1_obs")

    # allocate and extract shoebox
    strong_refls["shoebox"] = flex.shoebox(
        strong_refls["panel"], strong_refls["bbox"], allocate=True
    )
    strong_refls.extract_shoeboxes(experiment.imageset)

    _compute_mask(experiment, strong_refls, sigma_d, "s1_obs", strong_shoeboxes)

    logger.info(
        f"Computing background, intensity and centroids for {len(strong_refls)} reflections"
    )
    strong_refls.compute_background(experiments)
    strong_refls.compute_summed_intensity()

    n_sum = strong_refls.get_flags(strong_refls.flags.integrated_sum).count(True)
    logger.info(f"{n_sum} reflections integrated")

    strong_refls.compute_centroid(experiments)
    strong_refls.compute_d(experiments)

    return strong_refls, sigma_d


def final_integrator(
    experiments,
    reflection_table,
    sigma_d,
    use_crude_shoebox_mask=False,
    shoebox_probability=FULL_PARTIALITY,
):
    """Performs an initial integration of all predicted spots"""

    experiment = experiments[0]
    logger.info("\n" + "=" * 80 + "\nIntegrating reflections")

    # first compute the bbox
    if use_crude_shoebox_mask:
        reflection_table["bbox"] = _compute_bbox(
            experiment, reflection_table, sigma_d, "s1"
        )
    else:
        # compute bbox from model
        profile = experiment.crystal.mosaicity
        profile.parameterisation.compute_bbox(
            experiments, reflection_table, shoebox_probability
        )

    # Select reflections within detector
    x0, x1, y0, y1, _, _ = reflection_table["bbox"].parts()
    xsize, ysize = experiment.detector[0].get_image_size()
    selection = (x1 > 0) & (y1 > 0) & (x0 < xsize) & (y0 < ysize)
    reflection_table = reflection_table.select(selection)

    reflection_table["shoebox"] = flex.shoebox(
        reflection_table["panel"], reflection_table["bbox"], allocate=True
    )
    reflection_table.extract_shoeboxes(experiment.imageset)

    if use_crude_shoebox_mask:
        _compute_mask(experiment, reflection_table, sigma_d, "s1")
    else:
        # compute the mask from the model.
        profile = experiment.crystal.mosaicity
        profile.parameterisation.compute_mask(
            experiments, reflection_table, shoebox_probability
        )

    logger.info(
        f"Computing background, intensity, corrections for {len(reflection_table)} reflections"
    )
    reflection_table.compute_background(experiments)
    reflection_table.compute_summed_intensity()
    reflection_table.compute_corrections(experiments)
    reflection_table.compute_centroid(experiments)

    profile = experiment.crystal.mosaicity
    profile.parameterisation.compute_partiality(experiments, reflection_table)

    return reflection_table


def refine_profile(experiment, profile, refiner_data, wavelength_spread_model="delta"):
    """Do the profile refinement"""
    logger.info("\n" + "=" * 80 + "\nRefining profile parmameters")

    # Create the parameterisation
    state = ModelState(
        experiment,
        profile.parameterisation.parameterisation(),
        fix_orientation=True,
        fix_unit_cell=True,
        fix_wavelength_spread=wavelength_spread_model == "delta",
    )

    # Create the refiner and refine
    refiner = ProfileRefiner(state, refiner_data)
    refiner.refine()

    # Set the profile parameters
    profile.parameterisation.update_model(state)
    # Set the mosaicity
    experiment.crystal.mosaicity = profile

    return refiner


def refine_crystal(
    experiment,
    profile,
    refiner_data,
    fix_unit_cell=False,
    fix_orientation=False,
    wavelength_spread_model="delta",
):
    """Do the crystal refinement"""
    if (fix_unit_cell is True) and (fix_orientation is True):
        return

    logger.info("\n" + "=" * 80 + "\nRefining crystal parmameters")

    # Create the parameterisation
    state = ModelState(
        experiment,
        profile.parameterisation.parameterisation(),
        fix_mosaic_spread=True,
        fix_unit_cell=fix_unit_cell,
        fix_orientation=fix_orientation,
        fix_wavelength_spread=wavelength_spread_model == "delta",
    )

    # Create the refiner and refine
    refiner = ProfileRefiner(state, refiner_data)
    refiner.refine()

    return refiner


def predict_after_ellipsoid_refinement(experiment, reflection_table):
    """
    Predict the position of the spots

    """
    # Get some stuff from experiment
    A = np.array(experiment.crystal.get_A(), dtype=np.float64).reshape((3, 3))
    s0 = np.array([experiment.beam.get_s0()], dtype=np.float64).reshape(3, 1)
    s0_length = norm(s0)

    # Compute the vector to the reciprocal lattice point
    # since this is not on the ewald sphere, lets call it s2
    h = reflection_table["miller_index"]
    s1 = flex.vec3_double(len(h))
    s2 = flex.vec3_double(len(h))
    for i in range(len(reflection_table)):
        r = np.matmul(A, np.array([h[i]], dtype=np.float64).reshape(3, 1))
        s2_i = r + s0
        s2[i] = matrix.col(flumpy.from_numpy(s2_i))
        s1[i] = matrix.col(flumpy.from_numpy(s2_i * s0_length / norm(s2_i)))
    reflection_table["s1"] = s1
    reflection_table["s2"] = s2
    reflection_table["entering"] = flex.bool(len(h), False)

    # Compute the ray intersections
    xyzpx = flex.vec3_double()
    xyzmm = flex.vec3_double()
    for i in range(len(s2)):
        ss = s1[i]
        mm = experiment.detector[0].get_ray_intersection(ss)
        px = experiment.detector[0].millimeter_to_pixel(mm)
        xyzpx.append(px + (0,))
        xyzmm.append(mm + (0,))
    reflection_table["xyzcal.mm"] = xyzmm
    reflection_table["xyzcal.px"] = xyzpx
    return reflection_table


def compute_prediction_probability(experiment, reflection_table):

    # Get stuff from experiment
    s0 = np.array([experiment.beam.get_s0()], dtype=np.float64).reshape(3, 1)
    s0_length = norm(s0)
    profile = experiment.crystal.mosaicity

    # Loop through reflections
    min_p = None
    for s2 in reflection_table["s2"]:
        s2 = np.array(list(s2), dtype=np.float64).reshape(3, 1)
        s3 = s2 * s0_length / norm(s2)
        r = s2 - s0
        epsilon = s3 - s2
        sigma = profile.parameterisation.sigma_for_reflection(s0, r)
        sigma_inv = inv(sigma)
        d = np.matmul(np.matmul(epsilon.T, sigma_inv), epsilon)[0, 0]
        p = chisq_pdf(3, float(d))
        if min_p is None or p < min_p:
            min_p = p

    # Print some stuff
    logger.info(
        "Quantile required to predicted all observed reflections = %.5f" % (1 - min_p)
    )


def run_ellipsoid_refinement(
    experiments,
    reflection_table,
    sigma_d,
    profile_model="angular2",
    wavelength_model="delta",
    fix_unit_cell=False,
    fix_orientation=False,
    capture_progress=False,
    n_cycles=3,
):
    """Runs ellipsoid refinement on strong spots.

    Creates the necessary data needed, then runs cycles of profile and crystal
    refinement,"""

    output_data = {
        "refiner_output": {
            "history": [],
            "correlation": None,
            "labels": None,
        },
    }

    # Set the M params
    if not hasattr(experiments[0].crystal, "mosaicity"):
        profile = ProfileModelFactory.from_sigma_d(profile_model, sigma_d)
    else:
        profile = experiments[0].crystal.mosaicity

    # Construct the profile refiner data
    refiner_data = RefinerData.from_reflections(experiments[0], reflection_table)

    # Do the refinement
    for _ in range(n_cycles):
        # refine the profile
        refiner = refine_profile(
            experiments[0],
            profile,
            refiner_data,
            wavelength_spread_model=wavelength_model,
        )
        if capture_progress:
            # Save some data for plotting later.
            output_data["refiner_output"]["history"].append(refiner.history)
            output_data["refiner_output"]["correlation"] = refiner.correlation()
            output_data["refiner_output"]["labels"] = refiner.labels()

        # refine the crystal
        _ = refine_crystal(
            experiments[0],
            profile,
            refiner_data,
            fix_unit_cell=fix_unit_cell,
            fix_orientation=fix_orientation,
            wavelength_spread_model=wavelength_model,
        )

    experiments[0].profile = profile

    # Post process the reflections
    # Update predictions
    reflection_table = predict_after_ellipsoid_refinement(
        experiments[0], reflection_table
    )
    # Compute prob
    compute_prediction_probability(experiments[0], reflection_table)

    return experiments, reflection_table, output_data


def predict(experiments, d_min=None, prediction_probability=FULL_PARTIALITY):
    """Predict the reflections"""
    logger.info("\n" + "=" * 80 + "\nPredicting reflections")

    # Set a resolution range
    if d_min is None:
        s0 = experiments[0].beam.get_s0()
        d_min = experiments[0].detector.get_max_resolution(s0)

    # Create the index generator
    index_generator = IndexGenerator(
        experiments[0].crystal.get_unit_cell(),
        experiments[0].crystal.get_space_group().type(),
        d_min,
    )

    # Get an array of miller indices
    miller_indices_to_test = index_generator.to_array()
    logger.info("Generated %d miller indices" % len(miller_indices_to_test))

    # Get the covariance matrix
    profile = experiments[0].crystal.mosaicity

    reflection_table = profile.parameterisation.predict_reflections(
        experiments, miller_indices_to_test, prediction_probability
    )

    # Do the prediction
    reflection_table.compute_d(experiments)
    logger.info("Predicted %d reflections" % len(reflection_table))

    return reflection_table



from __future__ import annotations

from dxtbx.serialize import load
from scitbx import matrix

import dials.command_line.find_rotation_axis as dials_find_rotation_axis


def test_find_rotation_axis(dials_data, run_in_tmp_path):
    myd88 = dials_data("MyD88_processed", pathlib=True)
    dials_find_rotation_axis.run(
        args=[str(myd88 / "imported.expt"), str(myd88 / "strong.refl")]
    )
    assert run_in_tmp_path.joinpath("optimised.expt").is_file()
    assert run_in_tmp_path.joinpath("find_rotation_axis-histogram.png").is_file()
    assert run_in_tmp_path.joinpath("find_rotation_axis-projection.png").is_file()
    experiments = load.experiment_list(
        run_in_tmp_path / "optimised.expt", check_format=False
    )

    axis = matrix.col(experiments[0].goniometer.get_rotation_axis())
    expected = matrix.col((-0.627963, -0.778243, 0))

    assert axis.angle(expected) < 1e-6



from __future__ import annotations

from dials.algorithms.refinement.outlier_detection import CentroidOutlier
from dials.array_family import flex


class Tukey(CentroidOutlier):
    """Implementation of the CentroidOutlier class using Tukey's rule of thumb.
    That is values more than iqr_multiplier times the interquartile range from
    the quartiles are designed outliers. When x=1.5, this is Tukey's rule."""

    def __init__(
        self,
        cols=None,
        min_num_obs=20,
        separate_experiments=True,
        separate_panels=True,
        separate_images=False,
        block_width=None,
        nproc=1,
        iqr_multiplier=1.5,
    ):

        if cols is None:
            cols = ["x_resid", "y_resid", "phi_resid"]
        CentroidOutlier.__init__(
            self,
            cols=cols,
            min_num_obs=min_num_obs,
            separate_experiments=separate_experiments,
            block_width=block_width,
            nproc=nproc,
            separate_panels=separate_panels,
            separate_images=separate_images,
        )

        self._iqr_multiplier = iqr_multiplier

        return

    def _detect_outliers(self, cols):

        from scitbx.math import five_number_summary

        outliers = flex.bool(len(cols[0]), False)
        for col in cols:
            min_x, q1_x, med_x, q3_x, max_x = five_number_summary(col)
            iqr_x = q3_x - q1_x
            cut_x = self._iqr_multiplier * iqr_x
            outliers.set_selected(col > q3_x + cut_x, True)
            outliers.set_selected(col < q1_x - cut_x, True)

        return outliers
